#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Jamdict console app
Latest version can be found at https://github.com/neocl/jamdict

References:
    Python documentation:
        https://docs.python.org/
    PEP 0008 - Style Guide for Python Code
        https://www.python.org/dev/peps/pep-0008/
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2017, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import json

from chirptext import confirm, TextReport, Timer
from chirptext.cli import CLIApp, setup_logging

from jamdict import Jamdict
from jamdict import config
from jamdict import version_info

# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------


JMD_XML = config.get_file('JMDICT_XML')
KD2_XML = config.get_file('KD2_XML')
JMNE_XML = config.get_file('JMNEDICT_XML')
JMD_DB = config.get_file('JAMDICT_DB')

if os.path.isfile('logging.json'):
    setup_logging('logging.json', 'logs')
else:
    setup_logging(os.path.join(config.home_dir(), 'logging.json'), 'logs')


# -------------------------------------------------------------------------------
# Functions
# -------------------------------------------------------------------------------

def get_jam(cli, args):
    if not args.jdb:
        args.jdb = None
    if args.kd2 or args.jmne:
        cli.logger.warning("Jamdict database location: {}".format(args.jdb))
        cli.logger.warning("Kanjidic2 database location: {}".format(args.kd2))
        jmd = Jamdict(db_file=args.jdb, kd2_file=args.kd2,
                      jmd_xml_file=args.jmdxml, kd2_xml_file=args.kd2xml,
                      jmnedict_file=args.jmne, jmnedict_xml_file=args.jmnexml)
    else:
        cli.logger.debug("Using the same database for both JMDict and Kanjidic2")
        jmd = Jamdict(db_file=args.jdb,
                      kd2_file=args.jdb,
                      jmnedict_file=args.jdb,
                      jmd_xml_file=args.jmdxml,
                      kd2_xml_file=args.kd2xml,
                      jmnedict_xml_file=args.jmnexml)
    if jmd.kd2 is None:
        cli.logger.warning("Kanjidic2 database could not be found")
    return jmd


def import_data(cli, args):
    '''Import XML data into SQLite database'''
    rp = TextReport()
    t = Timer(report=rp)
    db_loc = os.path.abspath(os.path.expanduser(args.jdb))
    show_info(cli, args)
    # rp.print("Jamdict DB location        : {}".format(db_loc))
    # rp.print("JMDict XML file location   : {}".format(args.jmdxml))
    # rp.print("Kanjidic2 XML file location: {}".format(args.kd2xml))
    # rp.print("JMnedict XML file location : {}".format(args.jmnexml))
    jam = get_jam(cli, args)
    if args and (args.jdb or args.kd2):
        if os.path.isfile(db_loc):
            if not confirm("Database file exists. Do you want to overwite (This action cannot be undone! yes/no?) "):
                cli.logger.warning("Program aborted.")
                exit()
            else:
                os.unlink(db_loc)
        # perform input
        t.start("Creating Jamdict SQLite database. This process may take very long time ...")
        jam.import_data()
        t.stop()
    else:
        print("Database paths were not provided. Process aborted.")


def dump_result(results, report=None):
    if report is None:
        report = TextReport()
    if results.entries:
        report.print("=" * 40)
        report.print("Found entries")
        report.print("=" * 40)
        for e in results.entries:
            kj = ', '.join([k.text for k in e.kanji_forms])
            kn = ', '.join([k.text for k in e.kana_forms])
            report.print("Entry: {} | Kj:  {} | Kn: {}".format(e.idseq, kj, kn))
            report.print("-" * 20)
            for idx, s in enumerate(e.senses):
                report.print("{idx}. {s}".format(idx=idx + 1, s=s))
            report.print('')
    else:
        report.print("No dictionary entry was found.")
    if results.chars:
        report.print("=" * 40)
        report.print("Found characters")
        report.print("=" * 40)
        for c in results.chars:
            report.print("Char: {} | Strokes: {}".format(c, c.stroke_count))
            report.print("-" * 20)
            for rmg in c.rm_groups:
                report.print("Readings:", ", ".join([r.value for r in rmg.readings]))
                report.print("Meanings:", ", ".join([m.value for m in rmg.meanings if not m.m_lang or m.m_lang == 'en']))
        report.print('')
    else:
        report.print("No character was found.")
    if results.names:
        report.print("=" * 40)
        report.print("Found name entities")
        report.print("=" * 40)
        for e in results.names:
            kj = ', '.join([k.text for k in e.kanji_forms])
            kn = ', '.join([k.text for k in e.kana_forms])
            report.print("Names: {} | Kj:  {} | Kn: {}".format(e.idseq, kj, kn))
            report.print("-" * 20)
            for idx, s in enumerate(e.senses):
                report.print("{idx}. {s}".format(idx=idx + 1, s=s))
            report.print('')
    else:
        report.print("No name was found.")


def lookup(cli, args):
    '''Lookup words by kanji/kana'''
    jam = get_jam(cli, args)
    results = jam.lookup(args.query, strict_lookup=args.strict)
    report = TextReport(args.output)
    if args.format == 'json':
        report.print(json.dumps(results.to_json(),
                                ensure_ascii=args.ensure_ascii,
                                indent=args.indent if args.indent else None))
    else:
        if args.compact:
            report.print(results.text(separator='\n------\n', entry_sep='\n'))
        else:
            dump_result(results, report=report)


def file_status(file_path):
    real_path = os.path.abspath(os.path.expanduser(file_path))
    return '[NOT FOUND]' if not os.path.isfile(real_path) else '[OK]'


def show_info(cli, args):
    ''' Show jamdict configuration (data folder, configuration file location, etc.) '''
    output = TextReport(args.output) if 'output' in args else TextReport()
    output.print("Jamdict " + version_info.__version__)
    output.print(version_info.__description__)
    output.header("Basic configuration")
    output.print("JAMDICT_HOME        : {}".format(config.home_dir()))
    output.print("Config file location: {}".format(config._get_config_manager().locate_config()))
    output.header("Data files")
    output.print("Jamdict DB location: {} - {}".format(args.jdb, file_status(args.jdb)))
    output.print("JMDict XML file    : {} - {}".format(args.jmdxml, file_status(args.jmdxml)))
    output.print("KanjiDic2 XML file : {} - {}".format(args.kd2xml, file_status(args.kd2xml)))
    output.print("JMnedict XML file : {} - {}".format(args.jmnexml, file_status(args.jmnexml)))


def show_version(cli, args):
    ''' Show Jamdict version '''
    if args.verbose:
        print("Jamdict {v} - {d}".format(d=version_info.__description__, v=version_info.__version__))
    else:
        print("Jamdict {}".format(version_info.__version__))


# -------------------------------------------------------------------------------
# Main
# -------------------------------------------------------------------------------

def add_data_config(parser):
    parser.add_argument('-j', '--jmdxml', help='Path to JMdict XML file', default=JMD_XML)
    parser.add_argument('-k', '--kd2xml', help='Path to KanjiDic2 XML file', default=KD2_XML)
    parser.add_argument('-e', '--jmnexml', help='Path to JMnedict XML file', default=JMNE_XML)
    parser.add_argument('-J', '--jdb', help='Path to JMDict SQLite file', default=JMD_DB)
    parser.add_argument('-K', '--kd2', help='Path to KanjiDic2 SQLite file', default=None)
    parser.add_argument('-E', '--jmne', help='Path to JMnedict SQLite file', default=None)


def main():
    '''Main entry of jamtk
    '''
    app = CLIApp(desc='Jamdict toolkit', logger=__name__)
    add_data_config(app.parser)

    # import task
    import_task = app.add_task('import', func=import_data)
    add_data_config(import_task)    

    # show info
    info_task = app.add_task('info', func=show_info)
    info_task.add_argument('-o', '--output', help='Write information to a text file')
    add_data_config(info_task)

    # show version
    version_task = app.add_task('version', func=show_version)
    add_data_config(version_task)

    # look up task
    lookup_task = app.add_task('lookup', func=lookup)
    lookup_task.add_argument('query', help='kanji/kana')
    lookup_task.add_argument('-f', '--format', help='json or text')
    lookup_task.add_argument('--compact', action='store_true')
    lookup_task.add_argument('-s', '--strict', action='store_true')
    lookup_task.add_argument('--ensure_ascii', help='Force JSON dumps to ASCII only', action='store_true')
    lookup_task.add_argument('--indent', help='JSON default indent', default=2, type=int)
    lookup_task.add_argument('-o', '--output', help='Path to a file to output lookup result, leave blank to write to console standard output')
    lookup_task.set_defaults(func=lookup)
    add_data_config(lookup_task)

    # run app
    app.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Script for testing KRad module library
Latest version can be found at https://github.com/neocl/jamdict

References:
     RADKFILE/KRADFILE
        This project provides a decomposition of kanji into a number of visual elements
        or radicals to support software which provides a lookup service using kanji components.
        https://www.edrdg.org/krad/kradinf.html

@author: Le Tuan Anh <tuananh.ke@gmail.com>
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Le Tuan Anh <tuananh.ke@gmail.com>"
__copyright__ = "Copyright 2016, jamdict"
__license__ = "MIT"

########################################################################

import os
import logging
import unittest
from jamdict import config
from jamdict.jmdict import JMDictXMLParser
from jamdict.kanjidic2 import Kanjidic2XMLParser
from jamdict.krad import KRad

########################################################################

MY_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA = os.path.join(MY_DIR, 'data')
MINI_JMD = os.path.join(TEST_DATA, 'JMdict_mini.xml')
MINI_KD2 = os.path.join(TEST_DATA, 'kanjidic2_mini.xml')
TEST_DB = os.path.join(TEST_DATA, 'jamdict_test.db')


def getLogger():
    return logging.getLogger(__name__)


class TestConfig(unittest.TestCase):

    def test_config(self):
        cfg = config.read_config()
        self.assertIn('KD2_XML', cfg)
        self.assertTrue(config.get_file('KD2_XML'))
        getLogger().info("jamdict log file location: {}".format(config._get_config_manager().locate_config()))


class TestModels(unittest.TestCase):

    def test_read_krad(self):
        krad = KRad()
        self.assertEqual(krad.krad['㘅'], ['亅', '二', '口', '彳', '金'])
        self.assertEqual(krad.krad['𪚲'], ['乙', '勹', '月', '田', '亀'])
        self.assertEqual(krad.radk['龠'], {'籥', '鸙', '龢', '龠', '龡', '籲', '瀹', '龥', '禴', '鑰', '爚', '龣'})


########################################################################

if __name__ == "__main__":
    logging.getLogger('jamdict').setLevel(logging.DEBUG)
    unittest.main()



# -*- coding: utf-8 -*-

'''
JMDict in SQLite format
Latest version can be found at https://github.com/neocl/jamdict

References:
    Python documentation:
        https://docs.python.org/
    PEP 0008 - Style Guide for Python Code
        https://www.python.org/dev/peps/pep-0008/
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2017, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging

from puchikarui import Schema
from . import __version__ as JAMDICT_VERSION, __url__ as JAMDICT_URL
from .jmdict import Meta, JMDEntry, EntryInfo, Link, BibInfo, Audit, KanjiForm, KanaForm, Sense, SenseGloss, LSource


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

MY_FOLDER = os.path.dirname(os.path.abspath(__file__))
SCRIPT_FOLDER = os.path.join(MY_FOLDER, 'data')
JMDICT_SETUP_FILE = os.path.join(SCRIPT_FOLDER, 'setup_jmdict.sql')
JMDICT_VERSION = '1.08'
JMDICT_URL = 'http://www.csse.monash.edu.au/~jwb/edict.html'
SETUP_SCRIPT = '''INSERT INTO meta VALUES ('jmdict.version', '{jv}');
INSERT INTO meta VALUES ('jmdict.url', '{ju}');
INSERT INTO meta VALUES ('generator', 'jamdict');
INSERT INTO meta VALUES ('generator_version', '{gv}');
INSERT INTO meta VALUES ('generator_url', '{gu}');'''.format(
    jv=JMDICT_VERSION,
    ju=JMDICT_URL,
    gv=JAMDICT_VERSION,
    gu=JAMDICT_URL
)


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Models
# -------------------------------------------------------------------------------

class JMDictSchema(Schema):

    KEY_JMD_VER = "jmdict.version"
    KEY_JMD_URL = "jmdict.url"

    def __init__(self, data_source=":memory:", setup_script=None, setup_file=None, *args, **kwargs):
        super().__init__(data_source, setup_script=setup_script, setup_file=setup_file, *args, **kwargs)
        self.add_script(SETUP_SCRIPT)
        self.add_file(JMDICT_SETUP_FILE)
        # Meta
        self.add_table('meta', ['key', 'value'], proto=Meta).set_id('key')
        self.add_table('Entry', ['idseq'])
        self.add_table('Link', ['ID', 'idseq', 'tag', 'desc', 'uri'])
        self.add_table('Bib', ['ID', 'idseq', 'tag', 'text'])
        self.add_table('Etym', ['idseq', 'text'])
        self.add_table('Audit', ['idseq', 'upd_date', 'upd_detl'])
        # Kanji
        self.add_table('Kanji', ['ID', 'idseq', 'text'])
        self.add_table('KJI', ['kid', 'text'])
        self.add_table('KJP', ['kid', 'text'])
        # Kana
        self.add_table('Kana', ['ID', 'idseq', 'text', 'nokanji'])
        self.add_table('KNI', ['kid', 'text'])
        self.add_table('KNP', ['kid', 'text'])
        self.add_table('KNR', ['kid', 'text'])
        # Senses
        self.add_table('Sense', ['ID', 'idseq'])
        self.add_table('stagk', ['sid', 'text'])
        self.add_table('stagr', ['sid', 'text'])
        self.add_table('pos', ['sid', 'text'])
        self.add_table('xref', ['sid', 'text'])
        self.add_table('antonym', ['sid', 'text'])
        self.add_table('field', ['sid', 'text'])
        self.add_table('misc', ['sid', 'text'])
        self.add_table('SenseInfo', ['sid', 'text'])
        self.add_table('SenseSource', ['sid', 'text', 'lang', 'lstype', 'wasei'])
        self.add_table('dialect', ['sid', 'text'])
        self.add_table('SenseGloss', ['sid', 'lang', 'gend', 'text'])


class JMDictSQLite(JMDictSchema):

    def __init__(self, db_path, setup_script=None, setup_file=None, *args, **kwargs):
        super().__init__(db_path, setup_script=setup_script, setup_file=setup_file, *args, **kwargs)

    def update_meta(self, version, url, ctx=None):
        # create a default context if none was provided
        if ctx is None:
            with self.open(ctx) as ctx:
                return self.update_meta(version, url, ctx=ctx)
        # else (a context is provided)
        # version
        jv = ctx.meta.by_id(self.KEY_JMD_VER)
        if not jv:
            ctx.meta.insert(self.KEY_JMD_VER, version)
        else:
            jv.value = version
            ctx.meta.save(jv)
        # url
        ju = ctx.meta.by_id(self.KEY_JMD_URL)
        if not ju:
            ctx.meta.insert(self.KEY_JMD_URL, version)
        else:
            ju.value = url
            ctx.meta.save(ju)

    def search(self, query, ctx=None, **kwargs):
        # ensure context
        if ctx is None:
            with self.ctx() as ctx:
                return self.search(query, ctx=ctx)
        _is_wildcard_search = '_' in query or '@' in query or '%' in query
        if _is_wildcard_search:
            where = "idseq IN (SELECT idseq FROM Kanji WHERE text like ?) OR idseq IN (SELECT idseq FROM Kana WHERE text like ?) OR idseq IN (SELECT idseq FROM sense JOIN sensegloss ON sense.ID == sensegloss.sid WHERE text like ?)"
        else:
            where = "idseq IN (SELECT idseq FROM Kanji WHERE text == ?) OR idseq IN (SELECT idseq FROM Kana WHERE text == ?) OR idseq IN (SELECT idseq FROM sense JOIN sensegloss ON sense.ID == sensegloss.sid WHERE text == ?)"
        getLogger().debug(where)
        params = [query, query, query]
        try:
            if query.startswith('id#'):
                query_int = int(query[3:])
                if query_int >= 0:
                    getLogger().debug("Searching by ID: {}".format(query_int))
                    where = "idseq = ?"
                    params = [query_int]
        except Exception:
            pass
        # else (a context is provided)
        eids = self.Entry.select(where, params, ctx=ctx)
        entries = []
        for e in eids:
            entries.append(self.get_entry(e.idseq, ctx=ctx))
        return entries

    def get_entry(self, idseq, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as new_context:
                return self.get_entry(idseq, new_context)
        # else (a context is provided)
        # select entry & info
        entry = JMDEntry(idseq)
        # links, bibs, etym, audit ...
        dblinks = ctx.Link.select('idseq=?', (idseq,))
        dbbibs = ctx.Bib.select('idseq=?', (idseq,))
        dbetym = ctx.Etym.select('idseq=?', (idseq,))
        dbaudit = ctx.Audit.select('idseq=?', (idseq,))
        if dblinks or dbbibs or dbetym or dbaudit:
            entry.info = EntryInfo()
            if dblinks:
                for l in dblinks:
                    entry.info.links.append(Link(l.tag, l.desc, l.uri))
            if dbbibs:
                for b in dbbibs:
                    entry.info.bibinfo.append(BibInfo(b.tag, b.text))
            if dbetym:
                for e in dbetym:
                    entry.info.etym.append(e)
            if dbaudit:
                for a in dbaudit:
                    entry.info.audit.append(Audit(e.upd_date, e.upd_detl))
        # select kanji
        kanjis = ctx.Kanji.select('idseq=?', (idseq,))
        for dbkj in kanjis:
            kj = KanjiForm(dbkj.text)
            kjis = ctx.KJI.select('kid=?', (dbkj.ID,))
            for i in kjis:
                kj.info.append(i.text)
            kjps = ctx.KJP.select('kid=?', (dbkj.ID,))
            for p in kjps:
                kj.pri.append(p.text)
            entry.kanji_forms.append(kj)
        # select kana
        kanas = ctx.Kana.select('idseq=?', (idseq,))
        for dbkn in kanas:
            kn = KanaForm(dbkn.text, dbkn.nokanji)
            knis = ctx.KNI.select('kid=?', (dbkn.ID,))
            for i in knis:
                kn.info.append(i.text)
            knps = ctx.KNP.select('kid=?', (dbkn.ID,))
            for p in knps:
                kn.pri.append(p.text)
            knrs = ctx.KNR.select('kid=?', (dbkn.ID,))
            for r in knrs:
                kn.restr.append(r.text)
            entry.kana_forms.append(kn)
        # select senses
        senses = ctx.Sense.select('idseq=?', (idseq,))
        for dbs in senses:
            s = Sense()
            # stagk
            ks = ctx.stagk.select('sid=?', (dbs.ID,))
            for k in ks:
                s.stagk.append(k.text)
            # stagr
            rs = ctx.stagr.select('sid=?', (dbs.ID,))
            for r in rs:
                s.stagr.append(r.text)
            # pos
            ps = ctx.pos.select('sid=?', (dbs.ID,))
            for p in ps:
                s.pos.append(p.text)
            # xref
            xs = ctx.xref.select('sid=?', (dbs.ID,))
            for x in xs:
                s.xref.append(x.text)
            # antonym
            ans = ctx.antonym.select('sid=?', (dbs.ID,))
            for a in ans:
                s.antonym.append(a.text)
            # field
            fs = ctx.field.select('sid=?', (dbs.ID,))
            for f in fs:
                s.field.append(f.text)
            # misc
            ms = ctx.misc.select('sid=?', (dbs.ID,))
            for m in ms:
                s.misc.append(m.text)
            # SenseInfo
            sis = ctx.SenseInfo.select('sid=?', (dbs.ID,))
            for si in sis:
                s.info.append(si.text)
            # SenseSource
            lss = ctx.SenseSource.select('sid=?', (dbs.ID,))
            for ls in lss:
                s.lsource.append(LSource(ls.lang, ls.lstype, ls.wasei, ls.text))
            # dialect
            ds = ctx.dialect.select('sid=?', (dbs.ID,))
            for d in ds:
                s.dialect.append(d.text)
            # SenseGloss
            gs = ctx.SenseGloss.select('sid=?', (dbs.ID,))
            for g in gs:
                s.gloss.append(SenseGloss(g.lang, g.gend, g.text))
            entry.senses.append(s)
        return entry

    def insert_entries(self, entries, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as new_context:
                return self.insert_entries(entries, ctx=new_context)
        # else
        getLogger().debug("JMdict bulk insert {} entries".format(len(entries)))
        for entry in entries:
            self.insert_entry(entry, ctx)

    def insert_entry(self, entry, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as ctx:
                return self.insert_entry(entry, ctx=ctx)
        # else (a context is provided)
        self.Entry.insert(entry.idseq, ctx=ctx)
        # insert info
        if entry.info:
            # links
            for lnk in entry.info.links:
                ctx.Link.insert(entry.idseq, lnk.tag, lnk.desc, lnk.uri)
            # bibs
            for bib in entry.info.bibinfo:
                ctx.Bib.insert(entry.idseq, bib.tag, bib.text)
            # etym
            for e in entry.info.etym:
                ctx.Etym.insert(entry.idseq, e)
            # audit
            for a in entry.info.audit:
                ctx.Audit.insert(entry.idseq, a.upd_date, a.upd_detl)
        # insert kanji
        for kj in entry.kanji_forms:
            kjid = ctx.Kanji.insert(entry.idseq, kj.text)
            # KJI
            for kji in kj.info:
                ctx.KJI.insert(kjid, kji)
            # KJP
            for kjp in kj.pri:
                ctx.KJP.insert(kjid, kjp)
            pass
        # insert kana
        for kn in entry.kana_forms:
            knid = ctx.Kana.insert(entry.idseq, kn.text, kn.nokanji)
            # KNI
            for kni in kn.info:
                ctx.KNI.insert(knid, kni)
            # KNP
            for knp in kn.pri:
                ctx.KNP.insert(knid, knp)
            # KNR
            for knr in kn.restr:
                ctx.KNR.insert(knid, knr)
            pass
        # insert senses
        for s in entry.senses:
            sid = ctx.Sense.insert(entry.idseq)
            # stagk
            for sk in s.stagk:
                ctx.stagk.insert(sid, sk)
            # stagr
            for sr in s.stagr:
                ctx.stagr.insert(sid, sr)
            # pos
            for pos in s.pos:
                ctx.pos.insert(sid, pos)
            # xref
            for xr in s.xref:
                ctx.xref.insert(sid, xr)
            # antonym
            for a in s.antonym:
                ctx.antonym.insert(sid, a)
            # field
            for f in s.field:
                ctx.field.insert(sid, f)
            # misc
            for m in s.misc:
                ctx.misc.insert(sid, m)
            # SenseInfo
            for i in s.info:
                ctx.SenseInfo.insert(sid, i)
            # SenseSource
            for l in s.lsource:
                ctx.SenseSource.insert(sid, l.text, l.lang, l.lstype, l.wasei)
            # dialect
            for d in s.dialect:
                ctx.dialect.insert(sid, d)
            # SenseGloss
            for g in s.gloss:
                ctx.SenseGloss.insert(sid, g.lang, g.gend, g.text)



# -*- coding: utf-8 -*-

'''
Module for retrieving kanji components (i.e. radicals)
Latest version can be found at https://github.com/neocl/jamdict

This package uses the RADKFILE/KRADFILE[1] file.
These files are the property of the [Electronic Dictionary Research and Development Group][2], and are used in conformance with the Group's [licence][3].

[1]: http://www.edrdg.org/krad/kradinf.html
[2]: http://www.edrdg.org/
[3]: http://www.edrdg.org/edrdg/licence.html

References:
    JMDict website:
        http://www.csse.monash.edu.au/~jwb/edict.html
    Python documentation:
        https://docs.python.org/
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging
import threading
from collections import OrderedDict
from collections import defaultdict as dd
from lxml import etree

from chirptext import chio
from chirptext.sino import Radical

# ------------------------------------------------------------------------------
# Configuration
# ------------------------------------------------------------------------------
MY_FOLDER = os.path.dirname(os.path.abspath(__file__))
DATA_FOLDER = os.path.join(MY_FOLDER, 'data')
KRADFILE = os.path.join(DATA_FOLDER, 'kradfile-u.gz')
RADKFILE = os.path.join(DATA_FOLDER, 'radkfile.gz')

logger = logging.getLogger(__name__)


########################################################################

class KRad:
    def __init__(self, **kwargs):
        """ Kanji Radical management
        """
        self.__krad_map = None
        self.__radk_map = None
        self.__rads = {}
        self.lock = threading.Lock()

    def _build_krad_map(self):
        with self.lock:
            lines = chio.read_file(KRADFILE, mode='rt').splitlines()
            # build the krad map
            self.__krad_map = {}
            self.__radk_map = dd(set)
            for line in lines:
                if line.startswith("#"):
                    continue
                else:
                    parts = line.split(':', maxsplit=1)
                    if len(parts) == 2:
                        rads = [r.strip() for r in parts[1].split()]
                        char_literal = parts[0].strip()
                        self.__krad_map[char_literal] = rads
                        for rad in rads:
                            self.__radk_map[rad].add(char_literal)

    @property
    def radk(self):
        if self.__radk_map is None:
            self._build_krad_map()
        return self.__radk_map

    @property
    def krad(self):
        if self.__krad_map is None:
            self._build_krad_map()
        return self.__krad_map



# -*- coding: utf-8 -*-

'''
Python library for manipulating Jim Breen's JMdict and KanjiDic2
Latest version can be found at https://github.com/neocl/jamdict

This package uses the [EDICT][1] and [KANJIDIC][2] dictionary files.
These files are the property of the [Electronic Dictionary Research and Development Group][3], and are used in conformance with the Group's [licence][4].

[1]: http://www.csse.monash.edu.au/~jwb/edict.html
[2]: http://www.csse.monash.edu.au/~jwb/kanjidic.html
[3]: http://www.edrdg.org/
[4]: http://www.edrdg.org/edrdg/licence.html

References:
    JMDict website:
        http://www.csse.monash.edu.au/~jwb/edict.html
    Python documentation:
        https://docs.python.org/
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

from . import __version__ as version_info
from .__version__ import __author__, __email__, __copyright__, __maintainer__
from .__version__ import __credits__, __license__, __description__, __url__
from .__version__ import __version_major__, __version_long__, __version__, __status__

from .jmdict_sqlite import JMDictSQLite
from .kanjidic2_sqlite import KanjiDic2SQLite
from .util import Jamdict, JMDictXML, KanjiDic2XML
__all__ = ['Jamdict', 'JMDictSQLite', 'JMDictXML', 'KanjiDic2SQLite', 'KanjiDic2XML',
           "__version__", "__author__", "__description__", "__copyright__", "version_info"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test script for JMDict SQLite
Latest version can be found at https://github.com/neocl/jamdict

References:
    Python unittest documentation:
        https://docs.python.org/3/library/unittest.html
    Python documentation:
        https://docs.python.org/
    PEP 0008 - Style Guide for Python Code
        https://www.python.org/dev/peps/pep-0008/
    PEP 0257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
'''

# Copyright (c) 2017, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Le Tuan Anh <tuananh.ke@gmail.com>"
__copyright__ = "Copyright 2016, jamdict"
__license__ = "MIT"

########################################################################

import sys
import os
import unittest
import logging

from jamdict import Jamdict
from jamdict import JMDictXML
from jamdict import JMDictSQLite


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_DATA = os.path.join(TEST_DIR, 'data')
if not os.path.isdir(TEST_DATA):
    os.makedirs(TEST_DATA)
TEST_DB = os.path.join(TEST_DATA, 'test.db')
MINI_JMD = os.path.join(TEST_DATA, 'JMdict_mini.xml')


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Test cases
# -------------------------------------------------------------------------------

class TestJamdictSQLite(unittest.TestCase):

    db = JMDictSQLite(TEST_DB)
    xdb = JMDictXML.from_file(MINI_JMD)
    ramdb = JMDictSQLite(":memory:", auto_expand_path=False)

    @classmethod
    def setUpClass(cls):
        if os.path.isfile(TEST_DB):
            getLogger().info("Removing previous database file at {}".format(TEST_DB))
            os.unlink(TEST_DB)

    def test_xml2sqlite(self):
        print("Test JMDict: XML to SQLite")
        try:
            self.db.insert_entries(self.xdb)
        except:
            getLogger().exception("Error happened while inserting entries")
            raise
            pass
        entries = self.db.Entry.select()
        self.assertEqual(len(entries), len(self.xdb))
        # test select entry by id
        e = self.db.get_entry(1001710)
        ejson = e.to_json()
        self.assertEqual(ejson['kanji'][0]['text'], 'お菓子')
        getLogger().debug(e.to_json())

    def test_xml2ramdb(self):
        print("Testing XML to RAM")
        noe = len(self.xdb)
        with self.ramdb.ctx() as ctx:
            self.ramdb.insert_entries(self.xdb, ctx=ctx)
            self.assertEqual(len(self.ramdb.Entry.select(ctx=ctx)), noe)

    def test_import_function(self):
        print("Testing JMDict import function")
        jd = Jamdict(db_file=":memory:", jmd_xml_file=MINI_JMD, auto_config=False, auto_expand=False)
        jd.import_data()

    def test_search(self):
        print("Test searching JMDict SQLite")
        with self.ramdb.ds.open() as ctx:
            self.ramdb.insert_entries(self.xdb, ctx=ctx)
            entries = ctx.Entry.select()
            print(len(entries))
            # Search by kana
            es = self.ramdb.search('あの', ctx)
            self.assertEqual(len(es), 2)
            getLogger().info('あの: {}'.format('|'.join([str(x) for x in es])))
            # Search by kanji
            es = self.db.search('%子%', ctx, exact_match=False)
            self.assertEqual(len(es), 4)
            getLogger().info('%子%: {}'.format('|'.join([str(x) for x in es])))
            # search by meaning
            es = self.db.search('%confections%', ctx, exact_match=False)
            self.assertTrue(es)
            getLogger().info('%confections%: {}'.format('|'.join([str(x) for x in es])))


# -------------------------------------------------------------------------------
# Main
# -------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

'''
Japanese Multilingual Named Entity Dictionary (JMnedict) in SQLite format
Latest version can be found at https://github.com/neocl/jamdict

References:
    ENAMDICT/JMnedict - Japanese Proper Names Dictionary Files
        https://www.edrdg.org/enamdict/enamdict_doc.html

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2020, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging

from puchikarui import Schema
from . import __version__ as JAMDICT_VERSION, __url__ as JAMDICT_URL
from .jmdict import Meta, JMDEntry, KanjiForm, KanaForm, Translation, SenseGloss


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

MY_FOLDER = os.path.dirname(os.path.abspath(__file__))
SCRIPT_FOLDER = os.path.join(MY_FOLDER, 'data')
JMNEDICT_SETUP_FILE = os.path.join(SCRIPT_FOLDER, 'setup_jmnedict.sql')
JMNEDICT_VERSION = '1.08'
JMNEDICT_URL = 'https://www.edrdg.org/enamdict/enamdict_doc.html'
JMNEDICT_DATE = '2020-05-29'
JMNEDICT_SETUP_SCRIPT = '''INSERT INTO meta VALUES ('jmnedict.version', '{jv}');
INSERT INTO meta VALUES ('jmnedict.url', '{ju}');
INSERT INTO meta VALUES ('jmnedict.date', '{jud}');
INSERT INTO meta SELECT 'generator', 'jamdict' WHERE NOT EXISTS (SELECT 1 FROM meta WHERE key = 'generator');
INSERT INTO meta SELECT 'generator_version', '{gv}' WHERE NOT EXISTS (SELECT 1 FROM meta WHERE key = 'generator_version');
INSERT INTO meta SELECT 'generator_url', '{gu}' WHERE NOT EXISTS (SELECT 1 FROM meta WHERE key = 'generator_url');'''.format(
    jv=JMNEDICT_VERSION,
    ju=JMNEDICT_URL,
    jud=JMNEDICT_DATE,
    gv=JAMDICT_VERSION,
    gu=JAMDICT_URL
)


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Models
# -------------------------------------------------------------------------------

class JMNEDictSchema(Schema):

    def __init__(self, data_source=":memory:", setup_script=None, setup_file=None, *args, **kwargs):
        super().__init__(data_source, setup_script=setup_script, setup_file=setup_file, *args, **kwargs)
        self.add_script(JMNEDICT_SETUP_SCRIPT)
        self.add_file(JMNEDICT_SETUP_FILE)
        # Meta
        self.add_table('meta', ['key', 'value'], proto=Meta).set_id('key')
        self.add_table('NEEntry', ['idseq'])
        # Kanji
        self.add_table('NEKanji', ['ID', 'idseq', 'text'])
        # Kana
        self.add_table('NEKana', ['ID', 'idseq', 'text', 'nokanji'])
        # Translation (~Sense of JMdict)
        self.add_table('NETranslation', ['ID', 'idseq'])
        self.add_table('NETransType', ['tid', 'text'])
        self.add_table('NETransXRef', ['tid', 'text'])
        self.add_table('NETransGloss', ['tid', 'lang', 'gend', 'text'])


class JMNEDictSQLite(JMNEDictSchema):

    def __init__(self, db_path, setup_script=None, setup_file=None, *args, **kwargs):
        super().__init__(db_path, setup_script=setup_script, setup_file=setup_file, *args, **kwargs)

    def search_ne(self, query, ctx=None, **kwargs):
        # ensure context
        if ctx is None:
            with self.ctx() as ctx:
                return self.search(query, ctx=ctx)
        _is_wildcard_search = '_' in query or '@' in query or '%' in query
        if _is_wildcard_search:
            where = "idseq IN (SELECT idseq FROM NEKanji WHERE text like ?) OR idseq IN (SELECT idseq FROM NEKana WHERE text like ?) OR idseq IN (SELECT idseq FROM NETranslation JOIN NETransGloss ON NETranslation.ID == NETransGloss.tid WHERE NETransGloss.text like ?) OR idseq IN (SELECT idseq FROM NETranslation JOIN NETransType ON NETranslation.ID == NETransType.tid WHERE NETransType.text like ?)"
        else:
            where = "idseq IN (SELECT idseq FROM NEKanji WHERE text == ?) OR idseq IN (SELECT idseq FROM NEKana WHERE text == ?) OR idseq IN (SELECT idseq FROM NETranslation JOIN NETransGloss ON NETranslation.ID == NETransGloss.tid WHERE NETransGloss.text == ?) or idseq in (SELECT idseq FROM NETranslation JOIN NETransType ON NETranslation.ID == NETransType.tid WHERE NETransType.text == ?)"
        getLogger().debug(where)
        params = [query, query, query, query]
        try:
            if query.startswith('id#'):
                query_int = int(query[3:])
                if query_int >= 0:
                    getLogger().debug("Searching NE by ID: {}".format(query_int))
                    where = "idseq = ?"
                    params = [query_int]
        except Exception:
            pass
        # else (a context is provided)
        eids = self.NEEntry.select(where, params, ctx=ctx)
        entries = []
        for e in eids:
            entries.append(self.get_ne(e.idseq, ctx=ctx))
        return entries

    def get_ne(self, idseq, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as new_context:
                return self.get_entry(idseq, new_context)
        # else (a context is provided)
        # select entry & info
        entry = JMDEntry(idseq)
        # select kanji
        kanjis = ctx.NEKanji.select('idseq=?', (idseq,))
        for dbkj in kanjis:
            kj = KanjiForm(dbkj.text)
            entry.kanji_forms.append(kj)
        # select kana
        kanas = ctx.NEKana.select('idseq=?', (idseq,))
        for dbkn in kanas:
            kn = KanaForm(dbkn.text, dbkn.nokanji)
            entry.kana_forms.append(kn)
        # select senses
        senses = ctx.NETranslation.select('idseq=?', (idseq,))
        for dbs in senses:
            s = Translation()
            # name_type
            nts = ctx.NETransType.select('tid=?', (dbs.ID,))
            for nt in nts:
                s.name_type.append(nt.text)
            # xref
            xs = ctx.NETransXRef.select('tid=?', (dbs.ID,))
            for x in xs:
                s.xref.append(x.text)
            # SenseGloss
            gs = ctx.NETransGloss.select('tid=?', (dbs.ID,))
            for g in gs:
                s.gloss.append(SenseGloss(g.lang, g.gend, g.text))
            entry.senses.append(s)
        return entry

    def insert_name_entities(self, entries, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as new_context:
                return self.insert_name_entities(entries, ctx=new_context)
        # else
        for entry in entries:
            self.insert_name_entity(entry, ctx)

    def insert_name_entity(self, entry, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as ctx:
                return self.insert_name_entity(entry, ctx=ctx)
        # else (a context is provided)
        self.NEEntry.insert(entry.idseq, ctx=ctx)
        # insert kanji
        for kj in entry.kanji_forms:
            ctx.NEKanji.insert(entry.idseq, kj.text)
        # insert kana
        for kn in entry.kana_forms:
            ctx.NEKana.insert(entry.idseq, kn.text, kn.nokanji)
        # insert translations
        for s in entry.senses:
            tid = ctx.NETranslation.insert(entry.idseq)
            # insert name_type
            for nt in s.name_type:
                ctx.NETransType.insert(tid, nt)
            # xref
            for xr in s.xref:
                ctx.NETransXRef.insert(tid, xr)
            # Gloss
            for g in s.gloss:
                ctx.NETransGloss.insert(tid, g.lang, g.gend, g.text)



# -*- coding: utf-8 -*-

'''
Python library for manipulating Jim Breen's KanjiDic2
Latest version can be found at https://github.com/neocl/jamdict

This package uses the [EDICT][1] and [KANJIDIC][2] dictionary files.
These files are the property of the [Electronic Dictionary Research and Development Group][3], and are used in conformance with the Group's [licence][4].

[1]: http://www.csse.monash.edu.au/~jwb/edict.html
[2]: http://www.edrdg.org/kanjidic/kanjd2index.html
[3]: http://www.edrdg.org/
[4]: http://www.edrdg.org/edrdg/licence.html

References:
    JMDict website:
        http://www.csse.monash.edu.au/~jwb/edict.html
        http://www.edrdg.org/kanjidic/kanjd2index.html
    Python documentation:
        https://docs.python.org/
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging
from lxml import etree

from chirptext import chio
from chirptext.sino import Radical as KangxiRadical

from .krad import KRad

# ------------------------------------------------------------------------------
# Configuration
# ------------------------------------------------------------------------------

krad = KRad()


def getLogger():
    return logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Models
# ------------------------------------------------------------------------------

class KanjiDic2(object):

    def __init__(self, file_version, database_version, date_of_creation):
        """
<!ELEMENT header (file_version,database_version,date_of_creation)>
<!--
    The single header element will contain identification information
    about the version of the file
    -->
<!ELEMENT file_version (#PCDATA)>
<!--
    This field denotes the version of kanjidic2 structure, as more
    than one version may exist.
    -->
<!ELEMENT database_version (#PCDATA)>
<!--
    The version of the file, in the format YYYY-NN, where NN will be
    a number starting with 01 for the first version released in a
    calendar year, then increasing for each version in that year.
    -->
<!ELEMENT date_of_creation (#PCDATA)>
<!--
    The date the file was created in international format (YYYY-MM-DD).
-->"""
        self.file_version = file_version
        self.database_version = database_version
        self.date_of_creation = date_of_creation
        self.characters = []

    def __len__(self):
        return len(self.characters)

    def __getitem__(self, idx):
        return self.characters[idx]


class Character(object):
    """<!ELEMENT character (literal,codepoint, radical, misc, dic_number?, query_code?, reading_meaning?)*>"""
    def __init__(self):
        """

        """
        self.ID = None
        self.literal = ''  # <!ELEMENT literal (#PCDATA)> The character itself in UTF8 coding.
        self.codepoints = []  # <!ELEMENT codepoint (cp_value+)>
        self.radicals = []  # <!ELEMENT radical (rad_value+)>
        self.__canon_radical = None
        self.stroke_count = None  # first stroke_count in misc
        self.grade = None  # <misc>/<grade>
        self.stroke_miscounts = []  # <misc>/stroke_count[1:]
        self.variants = []  # <misc>/<variant>
        self.freq = None  # <misc>/<freq>
        self.rad_names = []  # <misc>/<rad_name> a list of strings
        self.jlpt = None  # <misc>/<jlpt>
        self.dic_refs = []  # DicRef[]
        self.query_codes = []  # QueryCode[]
        self.rm_groups = []  # reading_meaning groups
        self.nanoris = []  # a list of strings

    def __repr__(self):
        meanings = self.meanings(english_only=True)
        return "{l}:{sc}:{meanings}".format(l=self.literal, sc=self.stroke_count, meanings=','.join(meanings))

    def __str__(self):
        return self.literal

    def meanings(self, english_only=False):
        ''' Accumulate all meanings '''
        meanings = []
        for rm in self.rm_groups:
            for m in rm.meanings:
                if english_only and m.m_lang != '':
                    continue
                meanings.append(m.value)
        return meanings

    @property
    def components(self):
        if self.literal in krad.krad:
            return krad.krad[self.literal]
        else:
            return []

    @property
    def radical(self):
        if self.__canon_radical is None:
            for rad in self.radicals:
                if rad.rad_type == 'classical':
                    self.__canon_radical = KangxiRadical.kangxi()[rad.value]
        return self.__canon_radical

    def to_json(self):
        return {'literal': self.literal,
                'codepoints': [cp.to_json() for cp in self.codepoints],
                'radicals': [r.to_json() for r in self.radicals],
                'stroke_count': self.stroke_count,
                'grade': self.grade if self.grade else '',
                'stroke_miscounts': self.stroke_miscounts,
                'variants': [v.to_json() for v in self.variants],
                'freq': self.freq if self.freq else 0,
                'rad_names': self.rad_names,
                'jlpt': self.jlpt if self.jlpt else '',
                'dic_refs': [r.to_json() for r in self.dic_refs],
                'q_codes': [q.to_json() for q in self.query_codes],
                'rm': [rm.to_json() for rm in self.rm_groups],
                'nanoris': list(self.nanoris)}


class CodePoint(object):

    def __init__(self, cp_type='', value=''):
        """<!ELEMENT cp_value (#PCDATA)>
    <!--
    The cp_value contains the codepoint of the character in a particular
    standard. The standard will be identified in the cp_type attribute.
    -->
        """
        self.cid = None
        self.cp_type = cp_type
        self.value = value

    def __repr__(self):
        if self.r_type:
            return "({t}) {v}".format(t=self.cp_type, v=self.value)
        else:
            return self.value

    def __str__(self):
        return self.value

    def to_json(self):
        return {'type': self.cp_type, 'value': self.value}


class Radical(object):

    def __init__(self, rad_type='', value=''):
        """<!ELEMENT radical (rad_value+)>
        <!ELEMENT rad_value (#PCDATA)>
        <!--
        The radical number, in the range 1 to 214. The particular
        classification type is stated in the rad_type attribute.
        -->"""
        self.cid = None
        self.rad_type = rad_type
        self.value = value

    def __repr__(self):
        if self.rad_type:
            return "({t}) {v}".format(t=self.rad_type, v=self.value)
        else:
            return self.value

    def __str__(self):
        return self.value

    def to_json(self):
        return {'type': self.rad_type, 'value': self.value}


class Variant(object):

    def __init__(self, var_type='', value=''):
        """<!ELEMENT variant (#PCDATA)>
        <!--
        Either a cross-reference code to another kanji, usually regarded as a
        variant, or an alternative indexing code for the current kanji.
        The type of variant is given in the var_type attribute.
        -->
        <!ATTLIST variant var_type CDATA #REQUIRED>
        <!--
        The var_type attribute indicates the type of variant code. The current
        values are:
        jis208 - in JIS X 0208 - kuten coding
        jis212 - in JIS X 0212 - kuten coding
        jis213 - in JIS X 0213 - kuten coding
          (most of the above relate to "shinjitai/kyuujitai"
          alternative character glyphs)
        deroo - De Roo number - numeric
        njecd - Halpern NJECD index number - numeric
        s_h - The Kanji Dictionary (Spahn & Hadamitzky) - descriptor
        nelson_c - "Classic" Nelson - numeric
        oneill - Japanese Names (O'Neill) - numeric
        ucs - Unicode codepoint- hex
        --> """
        self.cid = None
        self.var_type = var_type
        self.value = value

    def __repr__(self):
        if self.var_type:
            return "({t}) {v}".format(t=self.var_type, v=self.value)
        else:
            return self.value

    def __str__(self):
        return self.value

    def to_json(self):
        return {'type': self.var_type, 'value': self.value}


class DicRef(object):

    def __init__(self, dr_type='', value='', m_vol='', m_page=''):
        """<!ELEMENT dic_ref (#PCDATA)>
    <!--
    Each dic_ref contains an index number. The particular dictionary,
    etc. is defined by the dr_type attribute.
    -->
<!ATTLIST dic_ref dr_type CDATA #REQUIRED>
    <!--
    The dr_type defines the dictionary or reference book, etc. to which
    dic_ref element applies. The initial allocation is:
      nelson_c - "Modern Reader's Japanese-English Character Dictionary",
        edited by Andrew Nelson (now published as the "Classic"
        Nelson).
      nelson_n - "The New Nelson Japanese-English Character Dictionary",
        edited by John Haig.
      halpern_njecd - "New Japanese-English Character Dictionary",
        edited by Jack Halpern.
      halpern_kkld - "Kanji Learners Dictionary" (Kodansha) edited by
        Jack Halpern.
      heisig - "Remembering The  Kanji"  by  James Heisig.
      gakken - "A  New Dictionary of Kanji Usage" (Gakken)
      oneill_names - "Japanese Names", by P.G. O'Neill.
      oneill_kk - "Essential Kanji" by P.G. O'Neill.
      moro - "Daikanwajiten" compiled by Morohashi. For some kanji two
        additional attributes are used: m_vol:  the volume of the
        dictionary in which the kanji is found, and m_page: the page
        number in the volume.
      henshall - "A Guide To Remembering Japanese Characters" by
        Kenneth G.  Henshall.
      sh_kk - "Kanji and Kana" by Spahn and Hadamitzky.
      sakade - "A Guide To Reading and Writing Japanese" edited by
        Florence Sakade.
      jf_cards - Japanese Kanji Flashcards, by Max Hodges and
        Tomoko Okazaki. (Series 1)
      henshall3 - "A Guide To Reading and Writing Japanese" 3rd
        edition, edited by Henshall, Seeley and De Groot.
      tutt_cards - Tuttle Kanji Cards, compiled by Alexander Kask.
      crowley - "The Kanji Way to Japanese Language Power" by
        Dale Crowley.
      kanji_in_context - "Kanji in Context" by Nishiguchi and Kono.
      busy_people - "Japanese For Busy People" vols I-III, published
        by the AJLT. The codes are the volume.chapter.
      kodansha_compact - the "Kodansha Compact Kanji Guide".
      maniette - codes from Yves Maniette's "Les Kanjis dans la tete" French adaptation of Heisig.
    -->"""
        self.cid = None
        self.dr_type = dr_type
        self.value = value
        self.m_vol = m_vol
        self.m_page = m_page

    def __repr__(self):
        if self.dr_type:
            return "({t}) {v}".format(t=self.dr_type, v=self.value)
        else:
            return self.value

    def __str__(self):
        return self.value

    def to_json(self):
        return {'type': self.dr_type,
                'value': self.value,
                "m_vol": self.m_vol,
                "m_page": self.m_page}


class QueryCode(object):

    def __init__(self, qc_type='', value='', skip_misclass=""):
        """<!ELEMENT query_code (q_code+)>
    <!--
    These codes contain information relating to the glyph, and can be used
    for finding a required kanji. The type of code is defined by the
    qc_type attribute.
    -->
<!ELEMENT q_code (#PCDATA)>
    <!--
    The q_code contains the actual query-code value, according to the
    qc_type attribute.
    -->
<!ATTLIST q_code qc_type CDATA #REQUIRED>
    <!--
    The qc_type attribute defines the type of query code. The current values
    are:
      skip -  Halpern's SKIP (System  of  Kanji  Indexing  by  Patterns)
        code. The  format is n-nn-nn.  See the KANJIDIC  documentation
        for  a description of the code and restrictions on  the
        commercial  use  of this data. [P]  There are also
        a number of misclassification codes, indicated by the
        "skip_misclass" attribute.
      sh_desc - the descriptor codes for The Kanji Dictionary (Tuttle
        1996) by Spahn and Hadamitzky. They are in the form nxnn.n,
        e.g.  3k11.2, where the  kanji has 3 strokes in the
        identifying radical, it is radical "k" in the SH
        classification system, there are 11 other strokes, and it is
        the 2nd kanji in the 3k11 sequence. (I am very grateful to
        Mark Spahn for providing the list of these descriptor codes
        for the kanji in this file.) [I]
      four_corner - the "Four Corner" code for the kanji. This is a code
        invented by Wang Chen in 1928. See the KANJIDIC documentation
        for  an overview of  the Four Corner System. [Q]

      deroo - the codes developed by the late Father Joseph De Roo, and
        published in  his book "2001 Kanji" (Bonjinsha). Fr De Roo
        gave his permission for these codes to be included. [DR]
      misclass - a possible misclassification of the kanji according
        to one of the code types. (See the "Z" codes in the KANJIDIC
        documentation for more details.)
    -->
<!ATTLIST q_code skip_misclass CDATA #IMPLIED>
    <!--
    The values of this attribute indicate the type if
    misclassification:
    - posn - a mistake in the division of the kanji
    - stroke_count - a mistake in the number of strokes
    - stroke_and_posn - mistakes in both division and strokes
    - stroke_diff - ambiguous stroke counts depending on glyph
S    --> """
        self.cid = None
        self.qc_type = qc_type
        self.value = value
        self.skip_misclass = skip_misclass

    def __repr__(self):
        if self.qc_type:
            return "({t}) {v}".format(t=self.qc_type, v=self.value)
        else:
            return self.value

    def __str__(self):
        return self.value

    def to_json(self):
        return {'type': self.qc_type, 'value': self.value, "skip_misclass": self.skip_misclass}


class RMGroup(object):

    def __init__(self, readings=None, meanings=None):
        """<!ELEMENT reading_meaning (rmgroup*, nanori*)>
        <!--
        The readings for the kanji in several languages, and the meanings, also
        in several languages. The readings and meanings are grouped to enable
        the handling of the situation where the meaning is differentiated by
        reading. [T1]
        -->
        <!ELEMENT rmgroup (reading*, meaning*)>
        """
        self.ID = None
        self.cid = None
        self.readings = readings if readings else []
        self.meanings = meanings if meanings else []

    def __repr__(self):
        return "R: {} | M: {}".format(
            ", ".join([r.value for r in self.readings]),
            ", ".join(m.value for m in self.meanings))

    def __str__(self):
        return repr(self)

    def to_json(self):
        return {'readings': [r.to_json() for r in self.readings],
                'meanings': [m.to_json() for m in self.meanings]}


class Reading(object):

    def __init__(self, r_type='', value='', on_type="", r_status=""):
        """<!ELEMENT reading (#PCDATA)>
        <!--
        The reading element contains the reading or pronunciation
        of the kanji.
        -->
        <!ATTLIST reading r_type CDATA #REQUIRED>
        <!--
        The r_type attribute defines the type of reading in the reading
        element. The current values are:
        pinyin - the modern PinYin romanization of the Chinese reading
        of the kanji. The tones are represented by a concluding
        digit. [Y]
        korean_r - the romanized form of the Korean reading(s) of the
        kanji.  The readings are in the (Republic of Korea) Ministry
        of Education style of romanization. [W]
        korean_h - the Korean reading(s) of the kanji in hangul.
        ja_on - the "on" Japanese reading of the kanji, in katakana.
        Another attribute r_status, if present, will indicate with
        a value of "jy" whether the reading is approved for a
        "Jouyou kanji".
        A further attribute on_type, if present,  will indicate with
        a value of kan, go, tou or kan'you the type of on-reading.
        ja_kun - the "kun" Japanese reading of the kanji, usually in
        hiragana.
        Where relevant the okurigana is also included separated by a
        ".". Readings associated with prefixes and suffixes are
        marked with a "-". A second attribute r_status, if present,
        will indicate with a value of "jy" whether the reading is
        approved for a "Jouyou kanji".
        -->
        <!ATTLIST reading on_type CDATA #IMPLIED>
        <!--
        See under ja_on above.
        -->
        <!ATTLIST reading r_status CDATA #IMPLIED>
        <!--
        See under ja_on and ja_kun above.
        -->"""
        self.gid = None
        self.r_type = r_type
        self.value = value
        self.on_type = on_type
        self.r_status = r_status

    def __repr__(self):
        if self.r_type:
            return "({t}) {v}".format(t=self.r_type, v=self.value)
        else:
            return self.value

    def __str__(self):
        return self.value

    def to_json(self):
        return {'type': self.r_type,
                'value': self.value,
                'on_type': self.on_type,
                'r_status': self.r_status}


class Meaning(object):

    def __init__(self, value='', m_lang=''):
        """<!ELEMENT meaning (#PCDATA)>
        <!--
        The meaning associated with the kanji.
        -->
        <!ATTLIST meaning m_lang CDATA #IMPLIED>
        <!--
        The m_lang attribute defines the target language of the meaning. It
        will be coded using the two-letter language code from the ISO 639-1
        standard. When absent, the value "en" (i.e. English) is implied. [{}]
        -->"""
        self.gid = None
        self.m_lang = m_lang
        self.value = value

    def __repr__(self):
        if self.m_lang:
            return "({l}) {v}".format(l=self.m_lang, v=self.value)
        else:
            return self.value

    def __str__(self):
        return self.value

    def to_json(self):
        return {'m_lang': self.m_lang, 'value': self.value}


class Kanjidic2XMLParser(object):
    '''JMDict XML parser
    '''

    def __init__(self):
        pass

    def get_attrib(self, a_tag, attr_name, default_value=''):
        if attr_name == 'xml:lang':
            attr_name = '''{http://www.w3.org/XML/1998/namespace}lang'''
        if attr_name in a_tag.attrib:
            return a_tag.attrib[attr_name]
        else:
            return default_value

    def parse_file(self, kd2_file_path):
        ''' Parse all characters from Kanjidic2 XML file
        '''
        actual_path = os.path.abspath(os.path.expanduser(kd2_file_path))
        getLogger().debug('Loading data from file: {}'.format(actual_path))

        with chio.open(actual_path, mode='rb') as kd2file:
            tree = etree.iterparse(kd2file)
            kd2 = None
            for event, element in tree:
                if event == 'end':
                    if element.tag == 'header':
                        kd2 = self.parse_header(element)
                        element.clear()
                    elif element.tag == 'character':
                        kd2.characters.append(self.parse_char(element))
                        element.clear()
            return kd2

    def parse_header(self, e):
        fv = None
        dbv = None
        doc = None
        for child in e:
            if child.tag == 'file_version':
                fv = child.text
            elif child.tag == 'database_version':
                dbv = child.text
            elif child.tag == 'date_of_creation':
                doc = child.text
        return KanjiDic2(fv, dbv, doc)

    def parse_char(self, e):
        char = Character()
        for child in e:
            if child.tag == 'literal':
                char.literal = child.text
            elif child.tag == 'codepoint':
                self.parse_codepoint(child, char)
            elif child.tag == 'radical':
                self.parse_radical(child, char)
            elif child.tag == 'misc':
                self.parse_misc(child, char)
            elif child.tag == 'dic_number':
                self.parse_dic_refs(child, char)
            elif child.tag == 'query_code':
                self.parse_query_code(child, char)
            elif child.tag == 'reading_meaning':
                self.parse_reading_meaning(child, char)
            else:
                getLogger().warning("Unknown tag in child: {}".format(child.tag))
        return char

    def parse_codepoint(self, e, char):
        for child in e:
            if child.tag == 'cp_value':
                cp = CodePoint(self.get_attrib(child, 'cp_type'), child.text)
                char.codepoints.append(cp)
            else:
                getLogger().warning("Unknown tag: {}".format(child.tag))

    def parse_radical(self, e, char):
        for child in e:
            if child.tag == 'rad_value':
                rad = Radical(self.get_attrib(child, "rad_type"), child.text)
                char.radicals.append(rad)
            else:
                getLogger().warning("Unknown tag: {}".format(child.tag))

    def parse_misc(self, e, char):
        for child in e:
            # grade?, stroke_count+, variant*, freq?, rad_name*,jlpt?
            if child.tag == 'grade':
                char.grade = child.text
            elif child.tag == 'stroke_count':
                if char.stroke_count is None:
                    char.stroke_count = int(child.text)
                else:
                    char.stroke_miscounts.append(int(child.text))
            elif child.tag == 'variant':
                v = Variant(self.get_attrib(child, "var_type"), child.text)
                char.variants.append(v)
            elif child.tag == 'freq':
                char.freq = child.text
            elif child.tag == 'rad_name':
                char.rad_names.append(child.text)
            elif child.tag == 'jlpt':
                char.jlpt = child.text
            else:
                getLogger().warning("Unknown tag: {}".format(child.tag))

    def parse_dic_refs(self, e, char):
        for child in e:
            if child.tag == 'dic_ref':
                dr_type = self.get_attrib(child, "dr_type")
                m_vol = self.get_attrib(child, "m_vol")
                m_page = self.get_attrib(child, "m_page")
                dr = DicRef(dr_type, child.text, m_vol, m_page)
                char.dic_refs.append(dr)
            else:
                getLogger().warning("Unknown tag: {}".format(child.tag))

    def parse_query_code(self, e, char):
        for child in e:
            if child.tag == "q_code":
                qc_type = self.get_attrib(child, "qc_type")
                skip_misclass = self.get_attrib(child, "skip_misclass")
                char.query_codes.append(QueryCode(qc_type, child.text, skip_misclass))
            else:
                getLogger().warning("Unknown tag: {}".format(child.tag))

    def parse_reading_meaning(self, e, char):
        for child in e:
            if child.tag == "nanori":
                char.nanoris.append(child.text)
            elif child.tag == "rmgroup":
                rmgroup = RMGroup()
                char.rm_groups.append(rmgroup)
                for grandchild in child:
                    if grandchild.tag == 'reading':
                        r_type = self.get_attrib(grandchild, "r_type")
                        on_type = self.get_attrib(grandchild, "on_type")
                        r_status = self.get_attrib(grandchild, "r_status")
                        r = Reading(r_type, grandchild.text, on_type, r_status)
                        rmgroup.readings.append(r)
                    elif grandchild.tag == 'meaning':
                        m = Meaning(grandchild.text, self.get_attrib(grandchild, "m_lang"))
                        rmgroup.meanings.append(m)
                    else:
                        getLogger().warning("Unknown tag: {}".format(grandchild.tag))
            else:
                getLogger().warning("Unknown tag: {}".format(child.tag))



# -*- coding: utf-8 -*-

'''
KanjiDic2 in SQLite format
Latest version can be found at https://github.com/neocl/jamdict

References:
    KANJIDIC2 project
        https://www.edrdg.org/wiki/index.php/KANJIDIC_Project 

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2017, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging
from puchikarui import Schema
from . import __version__ as JAMDICT_VERSION, __url__ as JAMDICT_URL
from .jmdict import Meta
from .kanjidic2 import Character, CodePoint, Radical, Variant, DicRef, QueryCode, RMGroup, Reading, Meaning

# ------------------------------------------------------------------------------
# Configuration
# ------------------------------------------------------------------------------

MY_FOLDER = os.path.dirname(os.path.abspath(__file__))
SCRIPT_FOLDER = os.path.join(MY_FOLDER, 'data')
KANJIDIC2_VERSION = '1.6'
KANJIDIC2_URL = 'https://www.edrdg.org/wiki/index.php/KANJIDIC_Project'
KANJIDIC2_DATE = 'April 2008'
KANJIDIC2_SETUP_FILE = os.path.join(SCRIPT_FOLDER, 'setup_kanjidic2.sql')
KANJIDIC2_SETUP_SCRIPT = '''
INSERT INTO meta VALUES ('kanjidic2.version', '{kdv}');
INSERT INTO meta VALUES ('kanjidic2.url', '{kdu}');
INSERT INTO meta VALUES ('kanjidic2.date', '{kdd}');
INSERT INTO meta SELECT 'generator', 'jamdict'
WHERE NOT EXISTS (SELECT 1 FROM meta WHERE key='generator');
INSERT INTO meta SELECT 'generator_version', '{gv}'
WHERE NOT EXISTS (SELECT 1 FROM meta WHERE key='generator_version');
INSERT INTO meta SELECT 'generator_url', '{gu}'
WHERE NOT EXISTS (SELECT 1 FROM meta WHERE key='generator_url');'''.format(
    kdv=KANJIDIC2_VERSION,
    kdu=KANJIDIC2_URL,
    kdd=KANJIDIC2_DATE,
    gv=JAMDICT_VERSION,
    gu=JAMDICT_URL
)


def getLogger():
    return logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Models
# ------------------------------------------------------------------------------

class KanjiDic2Schema(Schema):

    KEY_FILE_VER = 'kanjidic2.file_version'
    KEY_DB_VER = 'kanjidic2.database_version'
    KEY_CREATED_DATE = 'kanjidic2.date_of_creation'

    def __init__(self, data_source, setup_script=None, setup_file=None, *args, **kwargs):
        super().__init__(data_source, setup_script=setup_script, setup_file=setup_file, *args, **kwargs)
        self.add_file(KANJIDIC2_SETUP_FILE)
        self.add_script(KANJIDIC2_SETUP_SCRIPT)
        # Meta
        self.add_table('meta', ['key', 'value'], proto=Meta).set_id('key')
        self.add_table('character', ['ID', 'literal', 'stroke_count', 'grade', 'freq', 'jlpt'], proto=Character, alias="char").set_id('ID')
        self.add_table('codepoint', ['cid', 'cp_type', 'value'], proto=CodePoint)
        self.add_table('radical', ['cid', 'rad_type', 'value'], proto=Radical)
        self.add_table('stroke_miscount', ['cid', 'value'], alias="smc")
        self.add_table('variant', ['cid', 'var_type', 'value'], proto=Variant)
        self.add_table('rad_name', ['cid', 'value'])
        self.add_table('dic_ref', ['cid', 'dr_type', 'value', 'm_vol', 'm_page'], proto=DicRef)
        self.add_table('query_code', ['cid', 'qc_type', 'value', 'skip_misclass'], proto=QueryCode)
        self.add_table('nanori', ['cid', 'value'])
        self.add_table('rm_group', ['ID', 'cid'], proto=RMGroup, alias='rmg').set_id('ID')
        self.add_table('reading', ['gid', 'r_type', 'value', 'on_type', 'r_status'], proto=Reading)
        self.add_table('meaning', ['gid', 'value', 'm_lang'], proto=Meaning)


class KanjiDic2SQLite(KanjiDic2Schema):

    def __init__(self, db_path, setup_script=None, setup_file=None, *args, **kwargs):
        super().__init__(db_path, setup_script=setup_script, setup_file=setup_file, *args, **kwargs)

    def update_meta(self, file_version, database_version, date_of_creation, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as new_context:
                return self.update_meta(file_version, database_version, date_of_creation, new_context)
        # else
        # file_version
        fv = ctx.meta.by_id(self.KEY_FILE_VER)
        if not fv:
            ctx.meta.insert(self.KEY_FILE_VER, file_version)
        else:
            fv.value = file_version
            ctx.meta.save(fv)
        # database_version
        dv = ctx.meta.by_id(self.KEY_DB_VER)
        if not dv:
            ctx.meta.insert(self.KEY_DB_VER, database_version)
        else:
            dv.value = database_version
            ctx.meta.save(dv)
        # date_of_creation
        doc = ctx.meta.by_id(self.KEY_CREATED_DATE)
        if not doc:
            ctx.meta.insert(self.KEY_CREATED_DATE, date_of_creation)
        else:
            doc.value = date_of_creation
            ctx.meta.save(doc)

    def insert_chars(self, chars, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as ctx:
                return self.insert_chars(chars, ctx=ctx)
        # else
        for c in chars:
            self.insert_char(c, ctx=ctx)

    def insert_char(self, c, ctx=None):
        # ensure context
        if ctx is None:
            with self.ctx() as ctx:
                return self.insert_char(c, ctx=ctx)
        # else
        c.ID = ctx.character.save(c)
        # save codepoints
        for cp in c.codepoints:
            cp.cid = c.ID
            ctx.codepoint.save(cp)
        # radicals
        for r in c.radicals:
            r.cid = c.ID
            ctx.radical.save(r)
        # stroke_miscount
        for smc in c.stroke_miscounts:
            ctx.smc.insert(c.ID, smc)
        # variants
        for v in c.variants:
            v.cid = c.ID
            ctx.variant.save(v)
        # radnames
        for rn in c.rad_names:
            ctx.rad_name.insert(c.ID, rn)
        # dic_refs
        for dr in c.dic_refs:
            dr.cid = c.ID
            ctx.dic_ref.save(dr)
        # query_codes
        for qc in c.query_codes:
            qc.cid = c.ID
            ctx.query_code.save(qc)
        # nanoris
        for n in c.nanoris:
            ctx.nanori.insert(c.ID, n)
        # reading groups
        for rmg in c.rm_groups:
            rmg.cid = c.ID
            rmg.ID = ctx.rmg.save(rmg)
            # save readings inside
            for r in rmg.readings:
                r.gid = rmg.ID
                ctx.reading.save(r)
            # save meanings inside
            for m in rmg.meanings:
                m.gid = rmg.ID
                ctx.meaning.save(m)

    def get_char(self, literal, ctx=None):
        if ctx is None:
            with self.ctx() as ctx:
                return self.get_char(literal, ctx=ctx)
        # context was ensured
        c = ctx.char.select_single('literal=?', (literal,))
        if not c:
            getLogger().debug("character {} could not be found".format(literal))
            return None
        else:
            return self.char_by_id(c.ID, ctx)

    def char_by_id(self, cid, ctx=None):
        if ctx is None:
            with self.ctx() as ctx:
                return self.select_char(cid, ctx=ctx)
        # context was ensured
        c = ctx.char.by_id(cid)
        c.codepoints = ctx.codepoint.select('cid=?', (cid,))
        c.radicals = ctx.radical.select('cid=?', (cid,))
        for smc in ctx.smc.select('cid=?', (cid,)):
            c.stroke_miscounts.append(smc.value)
        c.variants = ctx.variant.select('cid=?', (cid,))
        for r in ctx.rad_name.select('cid=?', (cid,)):
            c.rad_names.append(r.value)
        c.dic_refs = ctx.dic_ref.select('cid=?', (cid,))
        c.query_codes = ctx.query_code.select('cid=?', (cid,))
        for n in ctx.nanori.select('cid=?', (cid,)):
            c.nanoris.append(n.value)
        c.rm_groups = ctx.rmg.select('cid=?', (cid,))
        for rmg in c.rm_groups:
            rmg.readings = ctx.reading.select('gid=?', (rmg.ID,))
            rmg.meanings = ctx.meaning.select('gid=?', (rmg.ID,))
        return c



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Script for testing jamdict library
Latest version can be found at https://github.com/neocl/jamdict

References:
    Python documentation:
        https://docs.python.org/
    Python unittest
        https://docs.python.org/3/library/unittest.html
    --
    argparse module:
        https://docs.python.org/3/howto/argparse.html
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Le Tuan Anh <tuananh.ke@gmail.com>"
__copyright__ = "Copyright 2016, jamdict"
__license__ = "MIT"

########################################################################

import os
import logging
import unittest
from jamdict import config
from jamdict.jmdict import JMDictXMLParser
from jamdict.kanjidic2 import Kanjidic2XMLParser
from jamdict import Jamdict, JMDictXML

########################################################################

MY_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA = os.path.join(MY_DIR, 'data')
MINI_JMD = os.path.join(TEST_DATA, 'JMdict_mini.xml')
MINI_KD2 = os.path.join(TEST_DATA, 'kanjidic2_mini.xml')
MINI_JMNE = os.path.join(TEST_DATA, 'jmendict_mini.xml')
TEST_DB = os.path.join(TEST_DATA, 'jamdict_test.db')


def getLogger():
    return logging.getLogger(__name__)


class TestConfig(unittest.TestCase):

    def test_config(self):
        cfg = config.read_config()
        self.assertIn('KD2_XML', cfg)
        self.assertTrue(config.get_file('KD2_XML'))
        getLogger().info("jamdict log file location: {}".format(config._get_config_manager().locate_config()))


class TestModels(unittest.TestCase):

    def test_basic_models(self):
        parser = JMDictXMLParser()
        entries = parser.parse_file(MINI_JMD)
        self.assertEqual(len(entries), 230)  # there are 230 test entries
        e = entries[0]
        self.assertEqual(len(e), 1)  # there is only 1 sense
        self.assertEqual(len(e[0].gloss), 1)  # there is only 1 sense
        # first sense in entry e to string -> with POS
        self.assertEqual(str(e[0]), 'repetition mark in katakana ((noun (common) (futsuumeishi)))')
        self.assertEqual(str(e[0].text()), 'repetition mark in katakana')  # compact is enabled by default
        self.assertEqual(str(e[0].gloss[0]), 'repetition mark in katakana')

    def test_lookup_result(self):
        jam = Jamdict(jmd_xml_file=MINI_JMD, kd2_xml_file=MINI_KD2, auto_config=False, auto_expand=False)
        result = jam.lookup('おみやげ')
        print(repr(result))
        self.assertTrue(result.entries)
        self.assertEqual(result.entries[0].kana_forms[0].text, 'おみやげ')
        # test lookup by ID
        res = jam.lookup('id#{}'.format(1002490))
        self.assertTrue(res.entries)
        self.assertEqual(res.entries[0].kana_forms[0].text, 'おとそ')


class TestJamdictXML(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if os.path.isfile(TEST_DB):
            os.unlink(TEST_DB)

    def test_jmdict_xml(self):
        print("Test JMDict - lookup from XML")
        parser = JMDictXMLParser()
        entries = parser.parse_file(MINI_JMD)
        jmd = JMDictXML(entries)
        self.assertTrue(jmd.lookup(u'おてんき'))

    def test_jmdict_fields(self):
        parser = JMDictXMLParser()
        entries = parser.parse_file(MINI_JMD)
        jmd = JMDictXML(entries)
        results = jmd.lookup(u'おてんき')
        print(results)

    def test_jmdict_json(self):
        print("Test JMDict - XML to JSON")
        # Load mini dict data
        jmd = JMDictXML.from_file(MINI_JMD)
        e = jmd[10]
        self.assertIsNotNone(e)
        self.assertTrue(e.to_json())
        self.assertTrue(jmd[-1].to_json())

    def test_kanjidic2_xml(self):
        print("Test KanjiDic2 XML")
        # test module read kanjidic XML
        parser = Kanjidic2XMLParser()
        kd2 = parser.parse_file(MINI_KD2)
        for c in kd2:
            self.assertIsNotNone(c)
            for g in c.rm_groups:
                self.assertIsNotNone(g)
                self.assertTrue(g.readings)
                self.assertTrue(g.meanings)

    def test_kanjidic2_json(self):
        print("Test KanjiDic2 XML to JSON")
        parser = Kanjidic2XMLParser()
        kd2 = parser.parse_file(MINI_KD2)
        for c in kd2:
            self.assertIsNotNone(c.to_json())

    def test_jamdict_xml(self):
        print("Test Jamdict search in XML files")
        jam = Jamdict(jmd_xml_file=MINI_JMD, kd2_xml_file=MINI_KD2, auto_config=False)
        result = jam.lookup('おみやげ')
        self.assertEqual(len(result.entries), 1)
        self.assertEqual(len(result.chars), 2)
        self.assertEqual({c.literal for c in result.chars}, {'土', '産'})


class TestJamdictSQLite(unittest.TestCase):

    def test_jamdict_sqlite_all(self):
        if os.path.isfile(TEST_DB):
            os.unlink(TEST_DB)
        jam = Jamdict(db_file=TEST_DB, kd2_file=TEST_DB, jmnedict_file=TEST_DB, jmd_xml_file=MINI_JMD, kd2_xml_file=MINI_KD2, jmnedict_xml_file=MINI_JMNE)
        # Lookup using XML
        result = jam.jmdict_xml.lookup('おみやげ')
        getLogger().debug("Results: {}".format(result))
        # Lookup using SQLite
        jam.import_data()
        # test lookup
        result = jam.lookup('おみやげ')
        print(result.entries)
        self.assertEqual(len(result.entries), 1)
        self.assertEqual(len(result.chars), 2)
        self.assertEqual({c.literal for c in result.chars}, {'土', '産'})


########################################################################

if __name__ == "__main__":
    logging.getLogger('jamdict').setLevel(logging.DEBUG)
    unittest.main()



# -*- coding: utf-8 -*-

'''
Jamdict configuration management

Latest version can be found at https://github.com/neocl/jamdict

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging

from chirptext import AppConfig
from chirptext.chio import read_file, write_file

# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

MY_DIR = os.path.dirname(__file__)
CONFIG_TEMPLATE = os.path.join(MY_DIR, 'data', 'config_template.json')
__jamdict_home = os.environ.get('JAMDICT_HOME', MY_DIR)
__app_config = AppConfig('jamdict', mode=AppConfig.JSON, working_dir=__jamdict_home)


def getLogger():
    return logging.getLogger(__name__)


def _get_config_manager():
    ''' Internal function for retrieving application config manager object
    Don't use this directly, use read_config() method instead
    '''
    return __app_config


def read_config():
    if not __app_config.config and not __app_config.locate_config():
        # need to create a config
        config_dir = os.path.expanduser('~/.jamdict/')
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        cfg_loc = os.path.join(config_dir, 'config.json')
        default_config = read_file(CONFIG_TEMPLATE)
        getLogger().warning("Jamdict configuration file could not be found. A new configuration file will be generated at {}".format(cfg_loc))
        getLogger().debug("Default config: {}".format(default_config))
        write_file(cfg_loc, default_config)
    # read config
    config = __app_config.config
    return config


def home_dir():
    ''' Find JAMDICT_HOME folder.
    if there is an environment variable that points to an existing directory
    (e.g. export JAMDICT_HOME=/home/user/jamdict)
    that folder will be used instead of the configured in jamdict JSON config file
    '''
    _config = read_config()
    # [2020-06-01] Allow JAMDICT_HOME to be overridden by environment variables
    if 'JAMDICT_HOME' in os.environ:
        _env_jamdict_home = os.path.abspath(os.path.expanduser(os.environ['JAMDICT_HOME']))
        if os.path.isdir(_env_jamdict_home):
            getLogger().debug("JAMDICT_HOME: {}".format(_env_jamdict_home))
            return _env_jamdict_home
    return _config.get('JAMDICT_HOME', __jamdict_home)


def data_dir():
    _config = read_config()
    _data_dir = _config.get('JAMDICT_DATA', '{JAMDICT_HOME}/data').format(JAMDICT_HOME=home_dir())
    return _data_dir


def get_file(file_key):
    _config = read_config()
    _data_dir = data_dir()
    _value = _config.get(file_key)
    return _value.format(JAMDICT_DATA=_data_dir) if _value else ''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test script for Jamcha SQLite
Latest version can be found at https://github.com/neocl/jamdict

References:
    Python unittest documentation:
        https://docs.python.org/3/library/unittest.html
    Python documentation:
        https://docs.python.org/
    PEP 0008 - Style Guide for Python Code
        https://www.python.org/dev/peps/pep-0008/
    PEP 0257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
'''

# Copyright (c) 2017, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Le Tuan Anh <tuananh.ke@gmail.com>"
__copyright__ = "Copyright 2017, jamdict"
__license__ = "MIT"

########################################################################

import os
import unittest
import logging

from jamdict import KanjiDic2SQLite
from jamdict import KanjiDic2XML


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_DATA = os.path.join(TEST_DIR, 'data')
if not os.path.isdir(TEST_DATA):
    os.makedirs(TEST_DATA)
TEST_DB = os.path.join(TEST_DATA, 'jamcha.db')
MINI_KD2 = os.path.join(TEST_DATA, 'kanjidic2_mini.xml')


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Test cases
# -------------------------------------------------------------------------------

class TestJamdictSQLite(unittest.TestCase):

    db = KanjiDic2SQLite(TEST_DB)
    ramdb = KanjiDic2SQLite(":memory:", auto_expand_path=False)
    xdb = KanjiDic2XML.from_file(MINI_KD2)

    @classmethod
    def setUpClass(cls):
        if os.path.isfile(TEST_DB):
            getLogger().info("Removing previous database file at {}".format(TEST_DB))
            os.unlink(TEST_DB)

    def test_xml2sqlite(self):
        print("Test KanjiDic2 - XML to SQLite DB in RAM")
        getLogger().info("Testing using {} test characters".format(len(self.xdb)))
        db = self.ramdb
        with db.ctx() as ctx:
            fv = self.xdb.kd2.file_version
            dv = self.xdb.kd2.database_version
            doc = self.xdb.kd2.date_of_creation
            db.update_meta(fv, dv, doc, ctx)
            metas = ctx.meta.select()
            getLogger().debug("KanjiDic2 meta: {}".format(metas))
            for c in self.xdb:
                db.insert_char(c, ctx)
                c2 = db.char_by_id(c.ID, ctx)
                getLogger().debug("c-xml", c.to_json())
                getLogger().debug("c-sqlite", c2.to_json())
                self.assertEqual(c.to_json(), c2.to_json())
            # test searching
            # by id
            c = ctx.char.select_single()
            c = db.char_by_id(c.ID, ctx=ctx)
            self.assertIsNotNone(c)
            self.assertTrue(c.rm_groups[0].readings)
            self.assertTrue(c.rm_groups[0].meanings)
            # by literal
            c = db.get_char('持', ctx=ctx)
            self.assertEqual(c.literal, '持')
            self.assertTrue(c.rm_groups[0].readings)
            self.assertTrue(c.rm_groups[0].meanings)


# -------------------------------------------------------------------------------
# Main
# -------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

'''
Basic APIs for accessing a parsed JMDict

Latest version can be found at https://github.com/neocl/jamdict

This package uses the [EDICT][1] and [KANJIDIC][2] dictionary files.
These files are the property of the [Electronic Dictionary Research and Development Group][3], and are used in conformance with the Group's [licence][4].

[1]: http://www.csse.monash.edu.au/~jwb/edict.html
[2]: http://www.csse.monash.edu.au/~jwb/kanjidic.html
[3]: http://www.edrdg.org/
[4]: http://www.edrdg.org/edrdg/licence.html

References:
    JMDict website:
        http://www.csse.monash.edu.au/~jwb/edict.html
    Python documentation:
        https://docs.python.org/
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging
import threading
from collections import defaultdict as dd
from collections import OrderedDict
from chirptext.deko import HIRAGANA, KATAKANA

from . import config
from .jmdict import JMDictXMLParser
from .krad import KRad
from .jmdict_sqlite import JMDictSQLite
from .kanjidic2 import Kanjidic2XMLParser
from .kanjidic2_sqlite import KanjiDic2SQLite
from .jmnedict_sqlite import JMNEDictSQLite


########################################################################

def getLogger():
    return logging.getLogger(__name__)


########################################################################

class LookupResult(object):

    def __init__(self, entries, chars, names=None):
        self.entries = entries if entries else []
        self.chars = chars if chars else []
        self.names = names if names else []

    def text(self, compact=True, entry_sep='。', separator=' | ', no_id=False, with_chars=True):
        output = []
        if self.entries:
            entry_txts = []
            for idx, e in enumerate(self.entries, start=1):
                entry_txt = e.text(compact=compact, separator=' ', no_id=no_id)
                entry_txts.append("#{}: {}".format(idx, entry_txt))
            output.append("[Entries]")
            output.append(entry_sep)
            output.append(entry_sep.join(entry_txts))
        elif not compact:
            output.append("No entries")
        if self.chars and with_chars:
            if compact:
                chars_txt = ', '.join(str(c) for c in self.chars)
            else:
                chars_txt = ', '.join(repr(c) for c in self.chars)
            if output:
                output.append(separator)
            output.append("[Chars]")
            output.append(entry_sep)
            output.append(chars_txt)
        if self.names:
            name_txts = []
            for idx, n in enumerate(self.names, start=1):
                name_txt = n.text(compact=compact, separator=' ', no_id=no_id)
                name_txts.append("#{}: {}".format(idx, name_txt))
            if output:
                output.append(separator)
            output.append("[Names]")
            output.append(entry_sep)
            output.append(entry_sep.join(name_txts))
        return "".join(output) if output else "Found nothing"

    def __repr__(self):
        return self.text(compact=True)

    def __str__(self):
        return self.text(compact=False)

    def to_json(self):
        return {'entries': [e.to_json() for e in self.entries],
                'chars': [c.to_json() for c in self.chars],
                'names': [n.to_json() for n in self.names]}


class JamdictSQLite(KanjiDic2SQLite, JMNEDictSQLite, JMDictSQLite):

    def __init__(self, data_source, setup_script=None, setup_file=None, *args, **kwargs):
        super().__init__(data_source, setup_script=setup_script, setup_file=setup_file, *args, **kwargs)


class Jamdict(object):

    def __init__(self, db_file=None, kd2_file=None,
                 jmd_xml_file=None, kd2_xml_file=None,
                 auto_config=True, auto_expand=True, reuse_ctx=True,
                 jmnedict_file=None, jmnedict_xml_file=None,
                 **kwargs):
        # file paths configuration
        self.auto_expand = auto_expand
        self.db_file = db_file if db_file else config.get_file('JAMDICT_DB') if auto_config else None
        self.kd2_file = kd2_file if kd2_file else config.get_file('JAMDICT_DB') if auto_config else None
        self.jmnedict_file = jmnedict_file if jmnedict_file else config.get_file('JAMDICT_DB') if auto_config else None
        if not self.db_file or not os.path.isfile(self.db_file):
            getLogger().warning("JAMDICT_DB could NOT be found. Searching will be extremely slow. Please run `python3 -m jamdict.tools import` first")
        if not self.kd2_file or not os.path.isfile(self.kd2_file):
            getLogger().warning("Kanjidic2 database could NOT be found. Searching will be extremely slow. Please run `python3 -m jamdict.tools import` first")
        self.jmd_xml_file = jmd_xml_file if jmd_xml_file else config.get_file('JMDICT_XML') if auto_config else None
        self.kd2_xml_file = kd2_xml_file if kd2_xml_file else config.get_file('KD2_XML') if auto_config else None
        self.jmnedict_xml_file = jmnedict_xml_file if jmnedict_xml_file else config.get_file('JMNEDICT_XML') if auto_config else None
        # data sources
        self._db_sqlite = None
        self._kd2_sqlite = None
        self._jmne_sqlite = None
        self._jmd_xml = None
        self._kd2_xml = None
        self._jmne_xml = None
        self.__krad_map = None
        self.reuse_ctx = reuse_ctx
        self.__jm_ctx = None
        try:
            if self.reuse_ctx and self.db_file and os.path.isfile(self.db_file):
                self.__jm_ctx = self.jmdict.ctx()
        except Exception:
            getLogger().warning("JMdict data could not be accessed.")

    def __del__(self):
        if self.__jm_ctx is not None:
            try:
                # try to close default SQLite context if needed
                self.__jm_ctx.close()
            except Exception:
                pass

    @property
    def db_file(self):
        return self.__db_file

    @db_file.setter
    def db_file(self, value):
        if self.auto_expand and value:
            self.__db_file = os.path.abspath(os.path.expanduser(value))
        else:
            self.__db_file = None

    @property
    def kd2_file(self):
        return self.__kd2_file

    @kd2_file.setter
    def kd2_file(self, value):
        if self.auto_expand and value:
            self.__kd2_file = os.path.abspath(os.path.expanduser(value))
        else:
            self.__kd2_file = None

    @property
    def jmnedict_file(self):
        return self.__jmnedict_file

    @jmnedict_file.setter
    def jmnedict_file(self, value):
        if self.auto_expand and value:
            self.__jmnedict_file = os.path.abspath(os.path.expanduser(value))
        else:
            self.__jmnedict_file = None

    @property
    def jmdict(self):
        if not self._db_sqlite and self.db_file:
            with threading.Lock():
                if (not self.kd2_file or self.kd2_file == self.db_file) and (not self.jmnedict_file or self.jmnedict_file == self.db_file):
                    # Use 1 DB for all
                    self._db_sqlite = JamdictSQLite(self.db_file, auto_expand_path=self.auto_expand)
                    self._kd2_sqlite = self._db_sqlite
                    self._jmne_sqlite = self._db_sqlite
                else:
                    # use separated db files
                    # this is kept for backward compatible, may change in future code revision
                    self._db_sqlite = JMDictSQLite(self.db_file, auto_expand_path=self.auto_expand)
        return self._db_sqlite

    @property
    def kd2(self):
        if self._kd2_sqlite is None:
            if self.kd2_file is not None:
                with threading.Lock():
                    self._kd2_sqlite = KanjiDic2SQLite(self.kd2_file, auto_expand_path=self.auto_expand)
            else:
                self._kd2_sqlite = self.jmdict
        return self._kd2_sqlite

    @property
    def jmnedict(self):
        ''' JM NE SQLite database access object '''
        if self._jmne_sqlite is None:
            if self.jmnedict_file is not None:
                with threading.Lock():
                    self._jmne_sqlite = JMNEDictSQLite(self.jmnedict_file, auto_expand_path=self.auto_expand)
            else:
                self._jmne_sqlite = self.jmdict
        return self._jmne_sqlite

    @property
    def jmdict_xml(self):
        if not self._jmd_xml and self.jmd_xml_file:
            with threading.Lock():
                getLogger().info("Loading JMDict from XML file at {}".format(self.jmd_xml_file))
                self._jmd_xml = JMDictXML.from_file(self.jmd_xml_file)
                getLogger().info("Loaded JMdict entries: {}".format(len(self._jmd_xml)))
        return self._jmd_xml

    @property
    def krad(self):
        ''' Kanji to radicals map '''
        if not self.__krad_map:
            with threading.Lock():
                self.__krad_map = KRad()
        return self.__krad_map.krad

    @property
    def radk(self):
        ''' Radical to kanji map '''
        if not self.__krad_map:
            with threading.Lock():
                self.__krad_map = KRad()
        return self.__krad_map.radk
    
    @property
    def kd2_xml(self):
        if not self._kd2_xml and self.kd2_xml_file:
            with threading.Lock():
                getLogger().info("Loading KanjiDic2 from XML file at {}".format(self.kd2_xml_file))
                self._kd2_xml = KanjiDic2XML.from_file(self.kd2_xml_file)
                getLogger().info("Loaded KanjiDic2 entries: {}".format(len(self._kd2_xml)))
        return self._kd2_xml

    @property
    def jmne_xml(self):
        if not self._jmne_xml and self.jmnedict_xml_file:
            with threading.Lock():
                getLogger().info("Loading JMnedict from XML file at {}".format(self.jmnedict_xml_file))
                self._jmne_xml = JMNEDictXML.from_file(self.jmnedict_xml_file)
                getLogger().info("Loaded JMnedict entries: {}".format(len(self._jmne_xml)))
        return self._jmne_xml

    def has_kd2(self):
        return self.db_file is not None or self.kd2_file is not None or self.kd2_xml_file is not None

    def has_jmne(self):
        ''' Check if current database has jmne support '''
        if self.jmnedict is not None:
            m = self.jmnedict.meta.select_single('key=?', ('jmnedict.version',))
            return m is not None and len(m.value) > 0
        return None

    def is_available(self):
        return (self.db_file is not None or self.jmd_xml_file is not None or
                self.kd2_file is not None or self.kd2_xml_file is not None or
                self.jmnedict_file is not None or self.jmnedict_xml_file is not None)

    def import_data(self):
        ''' Import JMDict and KanjiDic2 data from XML to SQLite '''
        if self.jmdict and self.jmdict_xml:
            getLogger().info("Importing JMDict data")
            self.jmdict.insert_entries(self.jmdict_xml)
        if self.kd2 is not None and self.kd2_xml:
            getLogger().info("Importing KanjiDic2 data")
            self.kd2.insert_chars(self.kd2_xml)
        if self.jmnedict is not None and self.jmne_xml:
            getLogger().info("Importing JMNEdict data")
            self.jmnedict.insert_name_entities(self.jmne_xml)

    def get_ne(self, idseq, ctx=None):
        ''' Get name entity by idseq in JMnedict '''
        if self.jmnedict is not None:
            return self.jmnedict.get_ne(idseq, ctx=ctx)
        elif self.jmnedict_xml_file:
            return self.jmne_xml.lookup(idseq)
        else:
            raise LookupError("There is no KanjiDic2 data source available")

    def get_char(self, literal, ctx=None):
        if self.kd2 is not None:
            return self.kd2.get_char(literal, ctx=ctx)
        elif self.kd2_xml:
            return self.kd2_xml.lookup(literal)
        else:
            raise LookupError("There is no KanjiDic2 data source available")

    def get_entry(self, idseq):
        if self.jmdict:
            return self.jmdict.get_entry(idseq)
        elif self.jmdict_xml:
            return self.jmdict_xml.lookup(idseq)[0]
        else:
            raise LookupError("There is no backend data available")

    def lookup(self, query, strict_lookup=False, lookup_chars=True, ctx=None, lookup_ne=True, **kwargs):
        ''' Search words and characters and return a LookupResult object.

        Keyword arguments:
        query --- Text to query, may contains wildcard characters
        exact_match --- use exact SQLite matching (==) instead of wildcard matching (LIKE)
        strict_lookup --- Only look up the Kanji characters in query (i.e. discard characters from variants)
        lookup_chars --- set lookup_chars to False to disable character lookup
        ctx --- Database access context, can be reused for better performance
        lookup_ne --- Lookup name-entities
        '''
        if not self.is_available():
            raise LookupError("There is no backend data available")
        elif not query:
            raise ValueError("Query cannot be empty")
        if ctx is None and self.reuse_ctx and self.__jm_ctx is not None:
            ctx = self.__jm_ctx
        # Lookup words
        entries = []
        chars = []
        names = []
        if self.jmdict is not None:
            entries = self.jmdict.search(query, ctx=ctx)
        elif self.jmdict_xml:
            entries = self.jmdict_xml.lookup(query)
        if lookup_chars and self.has_kd2():
            # lookup each character in query and kanji readings of each found entries
            chars_to_search = OrderedDict({c: c for c in query})
            if not strict_lookup and entries:
                # auto add characters from entries
                for e in entries:
                    for k in e.kanji_forms:
                        for c in k.text:
                            if c not in HIRAGANA and c not in KATAKANA:
                                chars_to_search[c] = c
            for c in chars_to_search:
                result = self.get_char(c, ctx=ctx)
                if result is not None:
                    chars.append(result)
        # lookup name-entities
        if lookup_ne and self.has_jmne():
            names = self.jmnedict.search_ne(query, ctx=ctx)
        # finish
        return LookupResult(entries, chars, names)


class JMDictXML(object):
    ''' JMDict API for looking up information in XML
    '''
    def __init__(self, entries):
        self.entries = entries
        self._seqmap = {}  # entryID - entryObj map
        self._textmap = dd(set)
        # compile map
        for entry in self.entries:
            self._seqmap[entry.idseq] = entry
            for kn in entry.kana_forms:
                self._textmap[kn.text].add(entry)
            for kj in entry.kanji_forms:
                self._textmap[kj.text].add(entry)

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, idx):
        return self.entries[idx]

    def lookup(self, a_query):
        if a_query in self._textmap:
            return tuple(self._textmap[a_query])
        elif a_query.startswith('id#'):
            entry_id = a_query[3:]
            if entry_id in self._seqmap:
                return (self._seqmap[entry_id],)
        # found nothing
        return ()

    @staticmethod
    def from_file(filename):
        parser = JMDictXMLParser()
        return JMDictXML(parser.parse_file(os.path.abspath(os.path.expanduser(filename))))


class JMNEDictXML(JMDictXML):
    pass


class KanjiDic2XML(object):

    def __init__(self, kd2):
        """
        """
        self.kd2 = kd2
        self.char_map = {}
        for char in self.kd2:
            if char.literal in self.char_map:
                getLogger().warning("Duplicate character entry: {}".format(char.literal))
            self.char_map[char.literal] = char

    def __len__(self):
        return len(self.kd2)

    def __getitem__(self, idx):
        return self.kd2[idx]

    def lookup(self, char):
        if char in self.char_map:
            return self.char_map[char]
        else:
            return None

    @staticmethod
    def from_file(filename):
        parser = Kanjidic2XMLParser()
        return KanjiDic2XML(parser.parse_file(filename))



# -*- coding: utf-8 -*-

'''
Python library for manipulating Jim Breen's JMdict
Latest version can be found at https://github.com/neocl/jamdict

This package uses the [EDICT][1] and [KANJIDIC][2] dictionary files.
These files are the property of the [Electronic Dictionary Research and Development Group][3], and are used in conformance with the Group's [licence][4].

[1]: http://www.csse.monash.edu.au/~jwb/edict.html
[2]: http://www.csse.monash.edu.au/~jwb/kanjidic.html
[3]: http://www.edrdg.org/
[4]: http://www.edrdg.org/edrdg/licence.html

References:
    JMDict website:
        http://www.csse.monash.edu.au/~jwb/edict.html
    Python documentation:
        https://docs.python.org/
    PEP 257 - Python Docstring Conventions:
        https://www.python.org/dev/peps/pep-0257/

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# Copyright (c) 2016, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

########################################################################

import os
import logging
from lxml import etree

from chirptext import chio

logger = logging.getLogger(__name__)

########################################################################


class JMDEntry(object):
    ''' Entries consist of kanji elements, reading elements,
    general information and sense elements. Each entry must have at
    least one reading element and one sense element. Others are optional.
    DTD <!ELEMENT entry (ent_seq, k_ele*, r_ele+, info?, sense+)>'''

    def __init__(self, idseq=''):
        # A unique numeric sequence number for each entry
        self.idseq = idseq     # ent_seq
        self.kanji_forms = []  # k_ele*  => KanjiForm[]
        self.kana_forms = []   # r_ele+  => KanaForm[]
        self.info = None       # info?   => EntryInfo
        self.senses = []       # sense+

    def __len__(self):
        return len(self.senses)

    def __getitem__(self, idx):
        return self.senses[idx]

    def set_info(self, info):
        if self.info:
            logging.warning("WARNING: multiple info tag")
        self.info = info

    def text(self, compact=True, separator=' ', no_id=False):
        tmp = []
        if not compact and not no_id:
            tmp.append('[id#%s]' % self.idseq)
        if self.kana_forms:
            tmp.append(self.kana_forms[0].text)
        if self.kanji_forms:
            tmp.append("({})".format(self.kanji_forms[0].text))
        if self.senses:
            tmp.append(':')
            if len(self.senses) == 1:
                tmp.append(self.senses[0].text(compact=compact))
            else:
                for sense, idx in zip(self.senses, range(len(self.senses))):
                    tmp.append('{i}. {s}'.format(i=idx + 1, s=sense.text(compact=compact)))
        return separator.join(tmp)

    def __repr__(self):
        return self.text(compact=True)

    def __str__(self):
        return self.text(compact=False)

    def to_json(self):
        ed = {'idseq': self.idseq,
              'kanji': [x.to_json() for x in self.kanji_forms],
              'kana': [x.to_json() for x in self.kana_forms],
              'senses': [x.to_json() for x in self.senses]}
        if self.info:
            ed['info'] = self.info.to_json()
        return ed


class KanjiForm(object):
    ''' The kanji element, or in its absence, the reading element, is
    the defining component of each entry.
    The overwhelming majority of entries will have a single kanji
    element associated with a word in Japanese. Where there are
    multiple kanji elements within an entry, they will be orthographical
    variants of the same word, either using variations in okurigana, or
    alternative and equivalent kanji. Common "mis-spellings" may be
    included, provided they are associated with appropriate information
    fields. Synonyms are not included; they may be indicated in the
    cross-reference field associated with the sense element.
    DTD <!ELEMENT k_ele (keb, ke_inf*, ke_pri*)>
    text --- a kanji written form of an entry, string
    info --- coded information field, a list of strings
    pri --- relative priority of the entry, a list of strings
    '''

    def __init__(self, text=''):
        '''This element will contain a word or short phrase in Japanese
        which is written using at least one non-kana character (usually kanji,
        but can be other characters). The valid characters are
        kanji, kana, related characters such as chouon and kurikaeshi, and
        in exceptional cases, letters from other alphabets.
        '''
        self.text = text  # <!ELEMENT keb (#PCDATA)>

        '''This is a coded information field related specifically to the
        orthography of the keb, and will typically indicate some unusual
        aspect, such as okurigana irregularity.'''
        self.info = []  # <!ELEMENT ke_inf (#PCDATA)>*

        '''This and the equivalent re_pri field are provided to record
        information about the relative priority of the entry,  and consist
        of codes indicating the word appears in various references which
        can be taken as an indication of the frequency with which the word
        is used. This field is intended for use either by applications which
        want to concentrate on entries of  a particular priority, or to
        generate subset files.
        The current values in this field are:
        - news1/2: appears in the "wordfreq" file compiled by Alexandre Girardi
        from the Mainichi Shimbun. (See the Monash ftp archive for a copy.)
        Words in the first 12,000 in that file are marked "news1" and words
        in the second 12,000 are marked "news2".
        - ichi1/2: appears in the "Ichimango goi bunruishuu", Senmon Kyouiku
        Publishing, Tokyo, 1998.  (The entries marked "ichi2" were
        demoted from ichi1 because they were observed to have low
        frequencies in the WWW and newspapers.)
        - spec1 and spec2: a small number of words use this marker when they
        are detected as being common, but are not included in other lists.
        - gai1/2: common loanwords, based on the wordfreq file.
        - nfxx: this is an indicator of frequency-of-use ranking in the
        wordfreq file. "xx" is the number of the set of 500 words in which
        the entry can be found, with "01" assigned to the first 500, "02"
        to the second, and so on. (The entries with news1, ichi1, spec1 and
        gai1 values are marked with a "(P)" in the EDICT and EDICT2
        files.)
        The reason both the kanji and reading elements are tagged is because
        on occasions a priority is only associated with a particular
        kanji/reading pair.'''
        self.pri = []  # <!ELEMENT ke_pri (#PCDATA)>*

    def set_text(self, text):
        if self.text:
            logging.warning("WARNING: duplicated text for k_ele")
        self.text = text

    def to_json(self):
        kjd = {'text': self.text}
        if self.info:
            kjd['info'] = self.info
        if self.pri:
            kjd['pri'] = self.pri
        return kjd

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.text


class KanaForm(object):
    '''<!ELEMENT r_ele (reb, re_nokanji?, re_restr*, re_inf*, re_pri*)>
    The reading element typically contains the valid readings
    of the word(s) in the kanji element using modern kanadzukai.
    Where there are multiple reading elements, they will typically be
    alternative readings of the kanji element. In the absence of a
    kanji element, i.e. in the case of a word or phrase written
    entirely in kana, these elements will define the entry.
    text --- a kana written form of an entry, string
    nokanji --- True means this entry cannot be regarded as a true reading of the kanji, boolean
    restr --- use to restrict the reading to a subset of the available kanji forms, list of string
    info --- coded information field, a list of strings
    pri --- relative priority of the entry, a list of strings
    '''

    def __init__(self, text='', nokanji=False):
        '''this element content is restricted to kana and related
        characters such as chouon and kurikaeshi. Kana usage will be
        consistent between the keb and reb elements; e.g. if the keb
        contains katakana, so too will the reb.'''
        self.text = text  # <!ELEMENT reb (#PCDATA)>

        '''This element, which will usually have a null value, indicates
        that the reb, while associated with the keb, cannot be regarded
        as a true reading of the kanji. It is typically used for words
        such as foreign place names, gairaigo which can be in kanji or
        katakana, etc.'''
        self.nokanji = nokanji  # <!ELEMENT re_nokanji (#PCDATA)>?

        '''This element is used to indicate when the reading only applies
        to a subset of the keb elements in the entry. In its absence, all
        readings apply to all kanji elements. The contents of this element
        must exactly match those of one of the keb elements.'''
        self.restr = []  # <!ELEMENT re_restr (#PCDATA)>*

        '''General coded information pertaining to the specific reading.
        Typically it will be used to indicate some unusual aspect of
        the reading.'''
        self.info = []  # <!ELEMENT re_inf (#PCDATA)>*

        '''See the comment on ke_pri above.'''
        self.pri = []  # <!ELEMENT re_pri (#PCDATA)>*

    def set_text(self, text):
        if self.text:
            logging.warning("WARNING: duplicated text for k_ele")
        self.text = text

    def to_json(self):
        knd = {'text': self.text,
               'nokanji': self.nokanji}
        if self.restr:
            knd['restr'] = self.restr
        if self.info:
            knd['info'] = self.info
        if self.pri:
            knd['pri'] = self.pri
        return knd

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.text


class EntryInfo(object):
    '''General coded information relating to the entry as a whole.
    DTD: <!ELEMENT info (links*, bibl*, etym*, audit*)>
    '''
    def __init__(self):
        self.links = []    # link* => Link[]
        self.bibinfo = []  # bibl* => BibInfo[]
        '''This field is used to hold information about the etymology
        of the kanji or kana parts of the entry. For gairaigo,
        etymological information may also be in the <lsource> element.'''
        self.etym = []     # <!ELEMENT etym (#PCDATA)>*
        self.audit = []    # audit* => Audit[]

    def to_json(self):
        return {'links': [x.to_json() for x in self.links],
                'bibinfo': [x.to_json() for x in self.bibinfo],
                'etym': self.etym,
                'audit': [x.to_json() for x in self.audit]}


class Link(object):
    '''This element holds details of linking information to
    entries in other electronic repositories. The link_tag will be
    coded to indicate the type of link (text, image, sound), the
    link_desc will provided a textual label for the link, and the
    link_uri contains the actual URI.
    <!ELEMENT links (link_tag, link_desc, link_uri)>'''
    def __init__(self, tag, desc, uri):
        self.tag = tag    # <!ELEMENT link_tag (#PCDATA)>
        self.desc = desc  # <!ELEMENT link_desc (#PCDATA)>
        self.uri = uri    # <!ELEMENT link_uri (#PCDATA)>

    def to_json(self):
        return {'tag': self.tag,
                'desc': self.desc,
                'uri': self.uri}


class BibInfo(object):
    '''Bibliographic information about the entry. The bib_tag will a
    coded reference to an entry in an external bibliographic database.
    The bib_txt field may be used for brief (local) descriptions.
    <!ELEMENT bibl (bib_tag?, bib_txt?)>
    <!ELEMENT bib_tag (#PCDATA)>
    <!ELEMENT bib_txt (#PCDATA)>
    '''
    def __init__(self, tag='', text=''):
        self.tag = tag
        self.text = text

    def set_tag(self, tag):
        if self.tag:
            logging.warning("WARNING: duplicate tag in bibinfo")
        self.tag = tag

    def set_text(self, text):
        if self.text:
            logging.warning("WARNING: duplicate text in bibinfo")
        self.text = text

    def to_json(self):
        return {'tag': self.tag, 'text': self.text}


class Audit(object):
    '''The audit element will contain the date and other information
    about updates to the entry. Can be used to record the source of
    the material.
    <!ELEMENT audit (upd_date, upd_detl)>'''
    def __init__(self, upd_date, upd_detl):
        self.upd_date = upd_date  # <!ELEMENT upd_date (#PCDATA)>
        self.upd_detl = upd_detl  # <!ELEMENT upd_detl (#PCDATA)>

    def to_json(self):
        return {'upd_date': self.upd_date, 'upd_detl': self.upd_detl}


class Sense(object):
    '''The sense element will record the translational equivalent
    of the Japanese word, plus other related information. Where there
    are several distinctly different meanings of the word, multiple
    sense elements will be employed.
    <!ELEMENT sense (stagk*, stagr*, pos*, xref*, ant*, field*, misc*, s_inf*, lsource*, dial*, gloss*, example*)>
    '''
    def __init__(self):
        '''These elements, if present, indicate that the sense is restricted
        to the lexeme represented by the keb and/or reb.'''
        self.stagk = []  # <!ELEMENT stagk (#PCDATA)>
        self.stagr = []  # <!ELEMENT stagr (#PCDATA)>

        '''Part-of-speech information about the entry/sense. Should use
        appropriate entity codes. In general where there are multiple senses
        in an entry, the part-of-speech of an earlier sense will apply to
        later senses unless there is a new part-of-speech indicated.'''
        self.pos = []   # <!ELEMENT pos (#PCDATA)>

        '''This element is used to indicate a cross-reference to another
        entry with a similar or related meaning or sense. The content of
        this element is typically a keb or reb element in another entry. In some
        cases a keb will be followed by a reb and/or a sense number to provide
        a precise target for the cross-reference. Where this happens, a JIS
        "centre-dot" (0x2126) is placed between the components of the
        cross-reference.
        <!ELEMENT xref (#PCDATA)*>'''
        self.xref = []  # xref

        '''This element is used to indicate another entry which is an
        antonym of the current entry/sense. The content of this element
        must exactly match that of a keb or reb element in another entry.'''
        self.antonym = []  # <!ELEMENT ant (#PCDATA)*>

        '''Information about the field of application of the entry/sense.
        When absent, general application is implied. Entity coding for
        specific fields of application.'''
        self.field = []  # <!ELEMENT field (#PCDATA)>

        '''This element is used for other relevant information about
        the entry/sense. As with part-of-speech, information will usually
        apply to several senses.'''
        self.misc = []  # <!ELEMENT misc (#PCDATA)>

        '''The sense-information elements provided for additional
        information to be recorded about a sense. Typical usage would
        be to indicate such things as level of currency of a sense, the
        regional variations, etc.'''
        self.info = []  # <!ELEMENT s_inf (#PCDATA)>

        self.lsource = []  # <!ELEMENT lsource (#PCDATA)>

        '''For words specifically associated with regional dialects in
        Japanese, the entity code for that dialect, e.g. ksb for Kansaiben.'''
        self.dialect = []  # <!ELEMENT dial (#PCDATA)>

        self.gloss = []  # <!ELEMENT gloss (#PCDATA | pri)*>

        '''The example elements provide for pairs of short Japanese and
        target-language phrases or sentences which exemplify the usage of the
        Japanese head-word and the target-language gloss. Words in example
        fields would typically not be indexed by a dictionary application.'''
        # It seems that this field is not used anymore!
        self.examples = []  # <!ELEMENT example (#PCDATA)>

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.text(compact=False)

    def text(self, compact=True):
        tmp = [str(x) for x in self.gloss]
        if not compact and self.pos:
            return '{gloss} ({pos})'.format(gloss='/'.join(tmp), pos=('(%s)' % '|'.join(self.pos)))
        else:
            return '/'.join(tmp)

    def to_json(self):
        sd = {}
        if self.stagk:
            sd['stagk'] = self.stagk
        if self.stagr:
            sd['stagr'] = self.stagr
        if self.pos:
            sd['pos'] = self.pos
        if self.xref:
            sd['xref'] = self.xref
        if self.antonym:
            sd['antonym'] = self.antonym
        if self.field:
            sd['field'] = self.field
        if self.misc:
            sd['misc'] = self.misc
        if self.info:
            sd['SenseInfo'] = self.info
        if self.lsource:
            sd['SenseSource'] = [x.to_json() for x in self.lsource]
        if self.dialect:
            sd['dialect'] = self.dialect
        if self.gloss:
            sd['SenseGloss'] = [x.to_json() for x in self.gloss]
        return sd


class Translation(Sense):
    ''' The trans element will record the translational equivalent
    of the Japanese name, plus other related information. (JMendict)
    <!ELEMENT trans (name_type*, xref*, trans_det*)>'''
    def __init__(self):
        super().__init__()
        self.name_type = []  # mapped to name_type*
        self.xref = []  # mapped to xref
        self.gloss = []  # mapped to trans_det

    def name_type_human(self):
        return [JMENDICT_TYPE_MAP[x] if x in JMENDICT_TYPE_MAP else x for x in self.name_type]

    def text(self, compact=True):
        tmp = [str(x) for x in self.gloss]
        types = "/".join(self.name_type) if compact else "/".join(self.name_type_human())
        return '{gloss} ({types})'.format(gloss='/'.join(tmp), types=types)

    def to_json(self):
        sd = super().to_json()
        sd['name_type'] = self.name_type
        return sd


class SenseGloss(object):
    '''Within each sense will be one or more "glosses", i.e.
        target-language words or phrases which are equivalents to the
        Japanese word. This element would normally be present, however it
        may be omitted in entries which are purely for a cross-reference.
        DTD: <!ELEMENT gloss (#PCDATA | pri)*>

        <!ATTLIST gloss xml:lang CDATA "eng">
        The xml:lang attribute defines the target language of the
        gloss. It will be coded using the three-letter language code from
        the ISO 639 standard. When absent, the value "eng" (i.e. English)
        is the default value.
        <!ATTLIST gloss g_gend CDATA #IMPLIED>
        The g_gend attribute defines the gender of the gloss (typically
        a noun in the target language. When absent, the gender is either
        not relevant or has yet to be provided.
        <!ELEMENT pri (#PCDATA)>
        These elements highlight particular target-language words which
        are strongly associated with the Japanese word. The purpose is to
        establish a set of target-language words which can effectively be
        used as head-words in a reverse target-language/Japanese relationship.'''
    def __init__(self, lang, gend, text):
        self.lang = lang
        self.gend = gend
        self.text = text

    def __repr__(self):
        return str(self)

    def __str__(self):
        tmp = [self.text]
        if self.lang and self.lang != 'eng':
            # lang = eng is trivial
            tmp.append('(lang:%s)' % self.lang)
        if self.gend:
            tmp.append('(gend:%s)' % self.gend)
        return ' '.join(tmp)

    def to_json(self):
        gd = {}
        if self.lang:
            gd['lang'] = self.lang
        if self.gend:
            gd['gend'] = self.gend
        if self.text:
            gd['text'] = self.text
        return gd


class LSource:
    '''This element records the information about the source
    language(s) of a loan-word/gairaigo. If the source language is other
    than English, the language is indicated by the xml:lang attribute.
    The element value (if any) is the source word or phrase.
    <!ATTLIST lsource xml:lang CDATA "eng">
    The xml:lang attribute defines the language(s) from which
    a loanword is drawn.  It will be coded using the three-letter language
    code from the ISO 639-2 standard. When absent, the value "eng" (i.e.
    English) is the default value. The bibliographic (B) codes are used.
    <!ATTLIST lsource ls_type CDATA #IMPLIED>
    The ls_type attribute indicates whether the lsource element
    fully or partially describes the source word or phrase of the
    loanword. If absent, it will have the implied value of "full".
    Otherwise it will contain "part".
    <!ATTLIST lsource ls_wasei CDATA #IMPLIED>
    The ls_wasei attribute indicates that the Japanese word
    has been constructed from words in the source language, and
    not from an actual phrase in that language. Most commonly used to
    indicate "waseieigo".'''

    def __init__(self, lang, lstype, wasei, text):
        self.lang = lang
        self.lstype = lstype
        self.wasei = wasei
        self.text = text

    def to_json(self):
        return {'lang': self.lang,
                'lstype': self.lstype,
                'wasei': self.wasei,
                'text': self.text}


JMENDICT_TYPES = (("surname", "family or surname"),
                  ("place", "place name"),
                  ("unclass", "unclassified name"),
                  ("company", "company name"),
                  ("product", "product name"),
                  ("work", "work of art, literature, music, etc. name"),
                  ("masc", "male given name or forename"),
                  ("fem", "female given name or forename"),
                  ("person", "full name of a particular person"),
                  ("given", "given name or forename, gender not specified"),
                  ("station", "railway station"),
                  ("organization", "organization name"),
                  ("ok", "old or irregular kana form"))
JMENDICT_TYPE_MAP = dict(JMENDICT_TYPES)
JMENDICT_TYPE_MAP_DECODE = {v: k for k, v in JMENDICT_TYPES}


class Meta(object):

    def __init__(self, key='', value=''):
        self.key = key
        self.value = value

    def __repr__(self):
        return "{{{}: {}}}".format(self.key, self.value)

    def __str__(self):
        return repr(self)


class JMDictXMLParser(object):
    '''JMDict XML parser
    '''

    def __init__(self):
        pass

    def parse_file(self, jmdict_file_path):
        ''' Parse JMDict_e.xml file and return a list of JMDEntry objects
        '''
        actual_path = os.path.abspath(os.path.expanduser(jmdict_file_path))
        logger.debug('Loading data from file: {}'.format(actual_path))

        with chio.open(actual_path, mode='rb') as jmfile:
            tree = etree.iterparse(jmfile)
            entries = []
            for event, element in tree:
                if event == 'end' and element.tag == 'entry':
                    entries.append(self.parse_entry_tag(element))
                    # and then we can clear the element to save memory
                    element.clear()
            return entries

    def parse_entry_tag(self, etag):
        '''Parse a lxml XML Node and generate a JMDEntry entry'''
        entry = JMDEntry()
        # parse ent_seq
        for child in etag:
            if child.tag == 'ent_seq':
                self.parse_ent_seq(child, entry)
            elif child.tag == 'k_ele':
                self.parse_k_ele(child, entry)
            elif child.tag == 'r_ele':
                self.parse_r_ele(child, entry)
            elif child.tag == 'info':
                self.parse_info(child, entry)
            elif child.tag == 'sense':
                self.parse_sense(child, entry)
            elif child.tag == 'trans':
                # JMendict support
                self.parse_ne_translation(child, entry)
            else:
                raise Exception("Invalid tag: %s" % child.tag)
        return entry

    def parse_ent_seq(self, seq_tag, entry):
        idseq = seq_tag.text
        if entry.idseq:
            raise Exception("WARNING: duplicated ent_seq tag")
        entry.idseq = idseq

    def get_single(self, tag_name, a_tag):
        children = a_tag.findall(tag_name)
        if len(children) == 0:
            return None
        elif len(children) > 1:
            raise Exception("There are %s %s tags in %s" % (len(children), tag_name, a_tag.tag))
        else:
            return children[0]

    def parse_k_ele(self, k_ele, entry):
        kr = KanjiForm()
        for child in k_ele:
            if child.tag == 'keb':
                kr.set_text(child.text)
            elif child.tag == 'ke_inf':
                kr.info.append(child.text)
            elif child.tag == 'ke_pri':
                kr.pri.append(child.text)
            else:
                raise Exception("WARNING: invalid tag %s in k_ele" % child.tag)
        # parse kebs
        entry.kanji_forms.append(kr)
        return kr

    def parse_r_ele(self, r_ele, entry):
        kr = KanaForm()
        for child in r_ele:
            if child.tag == 'reb':
                kr.set_text(child.text)
            elif child.tag == 're_nokanji':
                kr.nokanji = True
            elif child.tag == 're_restr':
                kr.restr.append(child.text)
            elif child.tag == 're_inf':
                kr.info.append(child.text)
            elif child.tag == 're_pri':
                kr.pri.append(child.text)
            else:
                raise Exception("WARNING: invalid tag %s in r_ele" % child.tag)
        # parse kebs
        entry.kana_forms.append(kr)
        return kr

    def parse_info(self, info_tag, entry):
        einfo = EntryInfo()
        for child in info_tag:
            if child.tag == 'links':
                self.parse_link(child, einfo)
            elif child.tag == 'bibl':
                self.parse_bibinfo(child, einfo)
            elif child.tag == 'etym':
                einfo.etym.append(child.text)
            elif child.tag == 'audit':
                self.parse_audit(child, einfo)
            else:
                raise Exception("WARNING: invalid tag in info tag (child.tag = %s)" % child.tag)
        entry.set_info(einfo)
        return einfo

    def parse_link(self, link_tag, entry_info):
        tag = self.get_single('link_tag', link_tag).text
        desc = self.get_single('link_desc', link_tag).text
        uri = self.get_single('link_uri', link_tag).text
        link = Link(tag, desc, uri)
        entry_info.links.append(link)
        return link

    def parse_bibinfo(self, bib_tag, entry_info):
        bib = BibInfo()
        for child in bib_tag:
            if child.tag == 'bib_tag':
                bib.set_tag(child.text)
            elif child.tag == 'bib_txt':
                bib.set_text(child.text)
            else:
                raise Exception("WARNING: invalid tag in bibinfo (child.tag = %s)" % child.tag)
        entry_info.bibinfo.append(bib)
        return bib

    def parse_ne_translation(self, trans_tag, entry):
        translation = Translation()
        for child in trans_tag:
            if child.tag == 'name_type':
                _name_type = JMENDICT_TYPE_MAP_DECODE[child.text] if child.text in JMENDICT_TYPE_MAP_DECODE else child.text
                translation.name_type.append(_name_type)
            elif child.tag == 'trans_det':
                # add sensegloss
                lang = self.get_attrib(trans_tag, 'xml:lang', default_value='eng')
                gloss = SenseGloss(lang=lang, gend='', text=child.text)
                translation.gloss.append(gloss)
            elif child.tag == 'xref':
                translation.xref.append(child.text)
            else:
                raise Exception("Invalid tag: {} in JMendict/trans tag".format(child.tag))
        entry.senses.append(translation)
        return translation

    def parse_sense(self, sense_tag, entry):
        sense = Sense()
        for child in sense_tag:
            if child.tag == 'stagk':
                sense.stagk.append(child.text)
            elif child.tag == 'stagr':
                sense.stagr.append(child.text)
            elif child.tag == 'pos':
                sense.pos.append(child.text)
            elif child.tag == 'xref':
                sense.xref.append(child.text)
            elif child.tag == 'ant':
                sense.antonym.append(child.text)
            elif child.tag == 'field':
                sense.field.append(child.text)
            elif child.tag == 'misc':
                sense.misc.append(child.text)
            elif child.tag == 's_inf':
                sense.info.append(child.text)
            elif child.tag == 'dial':
                sense.dialect.append(child.text)
            elif child.tag == 'example':
                sense.examples.append(child.text)
            elif child.tag == 'lsource':
                self.parse_lsource(child, sense)
            elif child.tag == 'gloss':
                self.parse_sensegloss(child, sense)
            else:
                raise Exception("WARNING: invalid tag in sense tag (child.tag = %s) content = %s" % (child.tag, etree.tostring(child)))
        entry.senses.append(sense)
        return sense

    def get_attrib(self, a_tag, attr_name, default_value=''):
        if attr_name == 'xml:lang':
            attr_name = '''{http://www.w3.org/XML/1998/namespace}lang'''
        if attr_name in a_tag.attrib:
            return a_tag.attrib[attr_name]
        else:
            return default_value

    def parse_sensegloss(self, gloss_tag, sense):
        lang = self.get_attrib(gloss_tag, 'xml:lang')
        gend = self.get_attrib(gloss_tag, 'g_gend')
        text = gloss_tag.text  # TODO: pri tag? raw text?
        gloss = SenseGloss(lang, gend, text)
        sense.gloss.append(gloss)
        return gloss

    def parse_lsource(self, lsource_tag, sense):
        lang = self.get_attrib(lsource_tag, 'xml:lang')
        lstype = self.get_attrib(lsource_tag, 'ls_type')
        wasei = self.get_attrib(lsource_tag, 'ls_wasei')
        lsource = LSource(lang, lstype, wasei, lsource_tag.text)
        sense.lsource.append(lsource)
        return lsource
